/*
FILE:     erf_iolib.C
*/
/*
VERSION:  0.904-beta
*/
/*
DATE:     8/12/2003
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2003 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <iostream.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "ndb_misclib.h"

#include "CifFileObj.h"
#include "erf_iolib.h"
#include "erf_iolib_global.h"
#include "erf_util.h"
#include "util.h"

#define tokenbinfile          "ndb_tokens.odb"
#define spgbinfile            "space_group.odb"
#define cifbinfile            "ndb_cif.odb"
#define rmsbinfile            "ndb_cif_rms.odb"
#define refinementbinfile     "refinement.odb"
#define dictbinfile           "mmcif_pdbx.odb"

DICFileObj *dict = NULL;

int MxTokens = 0;
ndb_token_format *NdbToken = NULL;
int MxJrnlTokens = 0;
ndb_token_format *JrnlToken = NULL;
int MxAtomTokens = 0;
atom_format *AtomInfo = NULL;
int MxCategory = 0;
CIF_CATEGORY *CifCategory = NULL;
int mxPrograms = 0;
CIF_PROGRAMS *programs = NULL;
int mxRmsRecords = 0;
NDBCIF_RMS_RECORDS *rmsRecords = NULL;
int Max_Refinement_Programs = 0;
REFMET_AUTHOR *Refinement_programs = NULL;

static int MxSpaceGroup = 0;
static SPACE_GROUP *SpaceGroup = NULL;

static int read_ndb_tokens();
static int read_spacegroup_cif(void);
static int ndb_tokenkey_compare(const void *token1, const void  *token2);
static int spacegroup_compare(const void *group1, const void *group2);
static void find_ndb_important_token_id(void);
static int read_cif_db();
static int read_cif_rms();
static int read_refinement_program();
static int ndb_category_compare(const void *a, const void *b);

int erf_lib_init()
{
       char binaryfile[300];
       sprintf(binaryfile, "%s/data/binary/%s\0", root_dir, dictbinfile);
       dict = new DICFileObj(binaryfile, READ_MODE);

       if (!read_ndb_tokens()) return FALSE;
       if (!read_cif_db()) return FALSE;
       if (!read_cif_rms()) return FALSE;
       if (!read_refinement_program()) return FALSE;
       return TRUE;
}

void erf_lib_close()
{
       if (dict) delete dict;
       dict = NULL;
}

int find_card_id(char *CardString)
{
       ndb_token_format tmpToken;

       strcpy(tmpToken.TokenName, CardString);
       ndb_token_format *token = (ndb_token_format *) bsearch(&tmpToken,
           NdbToken, MxTokens, sizeof(ndb_token_format), ndb_tokenkey_compare);
       if (token != NULL)
            return (((int) token - (int) NdbToken) / sizeof(ndb_token_format));
       else {
            int str_len = strlen(CardString);
            if (str_len == 0) return (-1);
            else if (CardString[str_len-1] >= '1' &&
                     CardString[str_len-1] <= '9') {
                 strncpy(tmpToken.TokenName, CardString, str_len-1);
                 tmpToken.TokenName[str_len-1] = 0;
                 token = (ndb_token_format *) bsearch(&tmpToken, NdbToken,
                     MxTokens, sizeof(ndb_token_format), ndb_tokenkey_compare);
                 if (token != NULL) return(((int) token - (int) NdbToken)
                             / sizeof(ndb_token_format));
                 else return (-1);
            } else return (-1);
       }
}

int find_jrnl_id(char *JRNLCard)
{
       ndb_token_format tmpToken;

       strcpy(tmpToken.TokenName, JRNLCard);
       ndb_token_format *token = (ndb_token_format *) bsearch(&tmpToken, JrnlToken,
               MxJrnlTokens, sizeof(ndb_token_format), ndb_tokenkey_compare);
       if (token != NULL) return(((int) token - (int) JrnlToken)
               / sizeof(ndb_token_format));
       else return (-1);
}

SPACE_GROUP *get_spacegroup_info(const char *spgname)
{
       if (SpaceGroup == NULL || MxSpaceGroup == 0) {
            if (!read_spacegroup_cif()) return NULL;
       }
       SPACE_GROUP tmp;
       strcpy(tmp.SpaceGroupName, spgname);
       SPACE_GROUP *exist = (SPACE_GROUP *) bsearch(&tmp, SpaceGroup,
             MxSpaceGroup, sizeof(SPACE_GROUP), spacegroup_compare);
       if (!exist) {
            ndb_log_message(NDB_MSG_ERR, "Space group %s is not found\n",
                         spgname);
            exit(1);
       }
       return exist;
}

int is_atom_card(const int card_id)
{
       for (int i = 0; i < MxAtomTokens; i++) {
            if (card_id == AtomInfo[i].NdbTokenId - 1) {
                 return (AtomInfo[i].Esd);
            }
       }
       return (FALSE);
}

int get_atom_card_id(const int card_id)
{
       int atom_id = 0;
       for (atom_id = 0; atom_id < MxAtomTokens; atom_id++) {
            if (card_id == AtomInfo[atom_id].NdbTokenId - 1) break;
       }
       if (atom_id == MxAtomTokens) atom_id = -1;
       return (atom_id);
}

CIF_CATEGORY *find_category(const char* category_name)
{
       CIF_CATEGORY category;
       strcpy(category.category, category_name);
       return ((CIF_CATEGORY *) bsearch(&category, CifCategory, MxCategory,
                 sizeof(CIF_CATEGORY), ndb_category_compare));
}

static int read_ndb_tokens()
{
       char binaryfile[300];
       sprintf(binaryfile, "%s/data/binary/%s\0", root_dir, tokenbinfile);
       CifFileObj *fobjR = new CifFileObj(binaryfile, READ_MODE);
       if (!fobjR) return FALSE;

       if (!fobjR->IsTablePresent("NDB_TOKENS", "ndb_token")) {
            delete fobjR; return FALSE;
       }
       ISTable *Table = fobjR->GetTablePtr("NDB_TOKENS", "ndb_token");
       MxTokens = Table->GetNumRows();
       NdbToken = (ndb_token_format *) calloc(MxTokens,
                              sizeof(ndb_token_format));
       int name_index = Table->GetColumnIndex("name");
       int seri_index = Table->GetColumnIndex("serial_no");
       int cono_index = Table->GetColumnIndex("continuation_no");
       int cont_index = Table->GetColumnIndex("continuation");
       int matr_index = Table->GetColumnIndex("matrix_field");
       int ndbo_index = Table->GetColumnIndex("ndb_only");
       int pdbc_index = Table->GetColumnIndex("pdb_continued_no_field");
       int row = 0;
       CifString string;
       for (row = 0; row < MxTokens; row++) {
            if (!Table->GetCell(string, name_index, row))
                 strcpy(NdbToken[row].TokenName, string.Text());

            if (!Table->GetCell(string, seri_index, row))
                 NdbToken[row].SeqField = atoi(string.Text());

            if (!Table->GetCell(string, cono_index, row))
                 NdbToken[row].ConField = atoi(string.Text());

            if (!Table->GetCell(string, cont_index, row))
                 NdbToken[row].ContinuedF = atoi(string.Text());

            if (!Table->GetCell(string, matr_index, row))
                 NdbToken[row].MatrixType = atoi(string.Text());

            if (!Table->GetCell(string, ndbo_index, row))
                 NdbToken[row].NdbOnly = atoi(string.Text());

            if (!Table->GetCell(string, pdbc_index, row))
                 NdbToken[row].PdbConField = atoi(string.Text());

            NdbToken[row].DisplayOrder = row;
       }
       qsort((void *) NdbToken, MxTokens, sizeof(ndb_token_format),
                    ndb_tokenkey_compare);

       if (!fobjR->IsTablePresent("NDB_TOKENS", "ndb_field")) {
            delete fobjR; return FALSE;
       }
       Table = fobjR->GetTablePtr("NDB_TOKENS", "ndb_field");
       int token_index = Table->GetColumnIndex("token_name");
       name_index = Table->GetColumnIndex("name");
       int type_index = Table->GetColumnIndex("type");
       int width_index = Table->GetColumnIndex("width");
       int prec_index = Table->GetColumnIndex("precision");
       int start_index = Table->GetColumnIndex("start_column");
       ndbo_index = Table->GetColumnIndex("ndb_only");
       int mixed_index = Table->GetColumnIndex("mixed_cases");
       CifString TokenName("XXX", 4);
       int tokenId = -1;
       int subrow = 0, nfield, i;
       for (row = 0; row < Table->GetNumRows(); row++) {
            Table->GetCell(string, token_index, row);
            if (string != TokenName) {
                 TokenName = string;
                 nfield = 0;
                 i = row;
                 do {
                      Table->GetCell(string, token_index, i);
                      if (string == TokenName) nfield++;
                      i++;
                 } while (i < Table->GetNumRows() && TokenName == string);
                 tokenId = find_card_id(TokenName.Text());
                 if (tokenId != -1) {
                      NdbToken[tokenId].NumField = nfield;
                      NdbToken[tokenId].FieldList = (field_format *)
                          calloc(NdbToken[tokenId].NumField, sizeof(field_format));
                 }
                 subrow = 0;
            }
            if (tokenId != -1) {
                 if (!Table->GetCell(string, name_index, row)) {
                      NdbToken[tokenId].FieldList[subrow].FieldLabel = (char *)
                          calloc(string.Length()+1, sizeof(char));
                      strcpy(NdbToken[tokenId].FieldList[subrow].FieldLabel, string.Text());
                 } else {
                      NdbToken[tokenId].FieldList[subrow].FieldLabel = (char *)
                          calloc(1, sizeof(char));
                      NdbToken[tokenId].FieldList[subrow].FieldLabel[0] = '\0';
                 }

                 if (!Table->GetCell(string, type_index, row))
                      NdbToken[tokenId].FieldList[subrow].FieldType = atoi(string.Text());

                 if (!Table->GetCell(string, width_index, row))
                      NdbToken[tokenId].FieldList[subrow].FieldWidth = atoi(string.Text());

                 if (!Table->GetCell(string, prec_index, row))
                      NdbToken[tokenId].FieldList[subrow].FieldPrec = atoi(string.Text());

                 if (!Table->GetCell(string, start_index, row))
                      NdbToken[tokenId].FieldList[subrow].FieldStcol = atoi(string.Text());
 
                 if (!Table->GetCell(string, ndbo_index, row))
                      NdbToken[tokenId].FieldList[subrow].NdbOnly = atoi(string.Text());

                 if (!Table->GetCell(string, mixed_index, row))
                      NdbToken[tokenId].FieldList[subrow].MixedCase = atoi(string.Text());
                 subrow++;
            }
       }

       if (!fobjR->IsTablePresent("NDB_TOKENS", "jrnl_token")) {
            delete fobjR; return FALSE;
       }
       Table = fobjR->GetTablePtr("NDB_TOKENS", "jrnl_token");
       MxJrnlTokens = Table->GetNumRows();
       JrnlToken = (ndb_token_format *) calloc(MxJrnlTokens,
                              sizeof(ndb_token_format));
       name_index = Table->GetColumnIndex("name");
       seri_index = Table->GetColumnIndex("serial_no");
       cono_index = Table->GetColumnIndex("continuation_no");
       cont_index = Table->GetColumnIndex("continuation");
       matr_index = Table->GetColumnIndex("matrix_field");
       ndbo_index = Table->GetColumnIndex("ndb_only");
       pdbc_index = Table->GetColumnIndex("pdb_continued_no_field");
       for (row = 0; row < MxJrnlTokens; row++) {
            if (!Table->GetCell(string, name_index, row))
                 strcpy(JrnlToken[row].TokenName, string.Text());

            if (!Table->GetCell(string, seri_index, row))
                 JrnlToken[row].SeqField = atoi(string.Text());

            if (!Table->GetCell(string, cono_index, row))
                 JrnlToken[row].ConField = atoi(string.Text());

            if (!Table->GetCell(string, cont_index, row))
                 JrnlToken[row].ContinuedF = atoi(string.Text());

            if (!Table->GetCell(string, matr_index, row))
                 JrnlToken[row].MatrixType = atoi(string.Text());

            if (!Table->GetCell(string, ndbo_index, row))
                 JrnlToken[row].NdbOnly = atoi(string.Text());

            if (!Table->GetCell(string, pdbc_index, row))
                 JrnlToken[row].PdbConField = atoi(string.Text());

            JrnlToken[row].DisplayOrder = row;
       }
       qsort((void *) JrnlToken, MxJrnlTokens, sizeof(ndb_token_format),
                    ndb_tokenkey_compare);

       if (!fobjR->IsTablePresent("NDB_TOKENS", "jrnl_field")) {
            delete fobjR; return FALSE;
       }
       Table = fobjR->GetTablePtr("NDB_TOKENS", "jrnl_field");
       token_index = Table->GetColumnIndex("token_name");
       name_index = Table->GetColumnIndex("name");
       type_index = Table->GetColumnIndex("type");
       width_index = Table->GetColumnIndex("width");
       prec_index = Table->GetColumnIndex("precision");
       start_index = Table->GetColumnIndex("start_column");
       ndbo_index = Table->GetColumnIndex("ndb_only");
       mixed_index = Table->GetColumnIndex("mixed_cases");
       TokenName.Clear();
       TokenName.Copy("XXX");
       tokenId = -1;
       subrow = 0;
       for (row = 0; row < Table->GetNumRows(); row++) {
            Table->GetCell(string, token_index, row);
            if (string != TokenName) {
                 TokenName = string;
                 nfield = 0;
                 i = row;
                 do {
                      Table->GetCell(string, token_index, i);
                      if (string == TokenName) nfield++;
                      i++;
                 } while (i < Table->GetNumRows() && TokenName == string);
                 tokenId = find_jrnl_id(TokenName.Text());
                 if (tokenId != -1) {
                      JrnlToken[tokenId].NumField = nfield;
                      JrnlToken[tokenId].FieldList = (field_format *)
                          calloc(JrnlToken[tokenId].NumField, sizeof(field_format));
                 }
                 subrow = 0;
            }
            if (tokenId != -1) {
                 if (!Table->GetCell(string, name_index, row)) {
                      JrnlToken[tokenId].FieldList[subrow].FieldLabel = (char *)
                           calloc(string.Length()+1, sizeof(char));
                      strcpy(JrnlToken[tokenId].FieldList[subrow].FieldLabel, string.Text());
                 } else {
                      JrnlToken[tokenId].FieldList[subrow].FieldLabel = (char *)
                           calloc(1, sizeof(char));
                      JrnlToken[tokenId].FieldList[subrow].FieldLabel[0] = '\0';
                 }

                 if (!Table->GetCell(string, type_index, row))
                      JrnlToken[tokenId].FieldList[subrow].FieldType = atoi(string.Text());

                 if (!Table->GetCell(string, width_index, row))
                      JrnlToken[tokenId].FieldList[subrow].FieldWidth = atoi(string.Text());

                 if (!Table->GetCell(string, prec_index, row))
                      JrnlToken[tokenId].FieldList[subrow].FieldPrec = atoi(string.Text());

                 if (!Table->GetCell(string, start_index, row))
                      JrnlToken[tokenId].FieldList[subrow].FieldStcol = atoi(string.Text());
 
                 if (!Table->GetCell(string, ndbo_index, row))
                      JrnlToken[tokenId].FieldList[subrow].NdbOnly = atoi(string.Text());

                 if (!Table->GetCell(string, mixed_index, row))
                      JrnlToken[tokenId].FieldList[subrow].MixedCase = atoi(string.Text());
                 subrow++;
            }
       }

       if (!fobjR->IsTablePresent("NDB_TOKENS", "atom_token")) {
            delete fobjR; return FALSE;
       }
       Table = fobjR->GetTablePtr("NDB_TOKENS", "atom_token");
       token_index = Table->GetColumnIndex("token_name");
       int regul_index = Table->GetColumnIndex("regular_or_aux");
       int esd_index = Table->GetColumnIndex("esd");
       seri_index = Table->GetColumnIndex("serial_atom");
       int pdb_index = Table->GetColumnIndex("pdb_token_name");
       int table_index = Table->GetColumnIndex("table_name");
       MxAtomTokens = Table->GetNumRows();
       AtomInfo = (atom_format *) calloc(MxAtomTokens, sizeof(atom_format));
       for (row = 0; row < Table->GetNumRows(); row++) {
            if (!Table->GetCell(string, token_index, row))
                 AtomInfo[row].NdbTokenId = find_card_id(string.Text()) + 1;

            if (!Table->GetCell(string, regul_index, row))
                 AtomInfo[row].AuxAtom = atoi(string.Text());

            if (!Table->GetCell(string, esd_index, row))
                 AtomInfo[row].Esd = atoi(string.Text());

            if (!Table->GetCell(string, seri_index, row))
                 AtomInfo[row].SeqAtom = atoi(string.Text());

            if (!Table->GetCell(string, pdb_index, row)) {
                 AtomInfo[row].PdbTitle = (char *) calloc(string.Length()+1, sizeof(char));
                 strcpy(AtomInfo[row].PdbTitle, string.Text());
            } else {
                 AtomInfo[row].PdbTitle = (char *) calloc(1, sizeof(char));
                 AtomInfo[row].PdbTitle[0] = '\0';
            }

            if (!Table->GetCell(string, table_index, row)) {
                 AtomInfo[row].TableName = (char *) calloc(string.Length()+1, sizeof(char));
                 strcpy(AtomInfo[row].TableName, string.Text());
            } else {
                 AtomInfo[row].TableName = (char *) calloc(1, sizeof(char));
                 AtomInfo[row].TableName[0] = '\0';
            }
       }
       delete fobjR;

       int AuxAtom;
       for (row = 0; row < MxAtomTokens; row++) {
            AuxAtom = 1;
            if (!AtomInfo[row].AuxAtom) AuxAtom |= AUX_ATOM;
            if (AtomInfo[row].SeqAtom) AuxAtom |= SEQ_ATOM;
            AtomInfo[row].Esd = AuxAtom;
       }
       find_ndb_important_token_id();

       return TRUE;
}

static int read_spacegroup_cif(void)
{
       char binaryfile[300];
       sprintf(binaryfile, "%s/data/binary/%s\0", root_dir, spgbinfile);
       CifFileObj *fobjR = new CifFileObj(binaryfile, READ_MODE);
       if (!fobjR) return FALSE;

       if (!fobjR->IsTablePresent("NDB_SPACE_GROUP", "ndb_space_group")) {
            delete fobjR; return FALSE;
       }

       ISTable *Table = fobjR->GetTablePtr("NDB_SPACE_GROUP", "ndb_space_group");
       int name_index = Table->GetColumnIndex("name");
       int num_index = Table->GetColumnIndex("operation_number");
       int m11_index = Table->GetColumnIndex("matrix[1][1]");
       int m12_index = Table->GetColumnIndex("matrix[1][2]");
       int m13_index = Table->GetColumnIndex("matrix[1][3]");
       int m21_index = Table->GetColumnIndex("matrix[2][1]");
       int m22_index = Table->GetColumnIndex("matrix[2][2]");
       int m23_index = Table->GetColumnIndex("matrix[2][3]");
       int m31_index = Table->GetColumnIndex("matrix[3][1]");
       int m32_index = Table->GetColumnIndex("matrix[3][2]");
       int m33_index = Table->GetColumnIndex("matrix[3][3]");
       int t11_index = Table->GetColumnIndex("trans[1][1]");
       int t12_index = Table->GetColumnIndex("trans[1][2]");
       int t21_index = Table->GetColumnIndex("trans[2][1]");
       int t22_index = Table->GetColumnIndex("trans[2][2]");
       int t31_index = Table->GetColumnIndex("trans[3][1]");
       int t32_index = Table->GetColumnIndex("trans[3][2]");
       MxSpaceGroup = 0;
       CifString SpaceGroupName("XXX", 4);
       CifString string;
       int i = 0, j, k, row = 0, num;
       for (i = 0; i < Table->GetNumRows(); i++) {
            Table->GetCell(string, name_index, i);
            if (string != SpaceGroupName) {
                 SpaceGroupName = string;
                 MxSpaceGroup++;
            }
       }
       SpaceGroup = (SPACE_GROUP *) calloc(MxSpaceGroup, sizeof(SPACE_GROUP));

       SpaceGroupName.Clear();
       SpaceGroupName.Copy("XXX");
       for (i = 0; i < Table->GetNumRows(); i++) {
            Table->GetCell(string, name_index, i);
            if (string != SpaceGroupName) {
                 SpaceGroupName = string;
                 num = 0;
                 j = i;
                 do {
                      Table->GetCell(string, name_index, j);
                      if (string == SpaceGroupName) {
                           num++;
                           j++;
                      }
                 } while (j < Table->GetNumRows() && string == SpaceGroupName);
                 SpaceGroup[row].SpaceDataNo = num;
                 SpaceGroup[row].SpaceData = (SPACE_DATA *) calloc(num,
                                    sizeof(SPACE_DATA));
                 strcpy(SpaceGroup[row].SpaceGroupName, SpaceGroupName.Text());
                 num = 0;
                 for (k = i; k < j; k++) {
                      if (!Table->GetCell(string, m11_index, k))
                           SpaceGroup[row].SpaceData[num].Data[0] = atoi(string.Text());
                      if (!Table->GetCell(string, m12_index, k))
                           SpaceGroup[row].SpaceData[num].Data[1] = atoi(string.Text());
                      if (!Table->GetCell(string, m13_index, k))
                           SpaceGroup[row].SpaceData[num].Data[2] = atoi(string.Text());
                      if (!Table->GetCell(string, m21_index, k))
                           SpaceGroup[row].SpaceData[num].Data[3] = atoi(string.Text());
                      if (!Table->GetCell(string, m22_index, k))
                           SpaceGroup[row].SpaceData[num].Data[4] = atoi(string.Text());
                      if (!Table->GetCell(string, m23_index, k))
                           SpaceGroup[row].SpaceData[num].Data[5] = atoi(string.Text());
                      if (!Table->GetCell(string, m31_index, k))
                           SpaceGroup[row].SpaceData[num].Data[6] = atoi(string.Text());
                      if (!Table->GetCell(string, m32_index, k))
                           SpaceGroup[row].SpaceData[num].Data[7] = atoi(string.Text());
                      if (!Table->GetCell(string, m33_index, k))
                           SpaceGroup[row].SpaceData[num].Data[8] = atoi(string.Text());
                      if (!Table->GetCell(string, t11_index, k))
                           SpaceGroup[row].SpaceData[num].Data[9] = atoi(string.Text());
                      if (!Table->GetCell(string, t12_index, k))
                           SpaceGroup[row].SpaceData[num].Data[10] = atoi(string.Text());
                      if (!Table->GetCell(string, t21_index, k))
                           SpaceGroup[row].SpaceData[num].Data[11] = atoi(string.Text());
                      if (!Table->GetCell(string, t22_index, k))
                           SpaceGroup[row].SpaceData[num].Data[12] = atoi(string.Text());
                      if (!Table->GetCell(string, t31_index, k))
                           SpaceGroup[row].SpaceData[num].Data[13] = atoi(string.Text());
                      if (!Table->GetCell(string, t32_index, k))
                           SpaceGroup[row].SpaceData[num].Data[14] = atoi(string.Text());
                      num++;
                 }
                 row++;
                 i = j - 1;
            }
       }
       qsort((void *) SpaceGroup, MxSpaceGroup, sizeof(SPACE_GROUP), spacegroup_compare);

       if (fobjR) delete fobjR;

       return TRUE;
}

static int ndb_tokenkey_compare(const void *token1, const void *token2)
{
       return (strcmp(((ndb_token_format *) token1)->TokenName, 
                      ((ndb_token_format *) token2)->TokenName));
}

static int spacegroup_compare(const void *group1, const void *group2)
{
       return (strcmp(((SPACE_GROUP  *) group1)->SpaceGroupName,
                      ((SPACE_GROUP  *) group2)->SpaceGroupName));
}

static void find_ndb_important_token_id(void)
{
       ATOMN_TOKEN  = find_card_id("ATOMN")+1;
       ATOMP_TOKEN  = find_card_id("ATOMP")+1;
       ATOMS_TOKEN  = find_card_id("ATOMS")+1;
       AUTHOR_TOKEN = find_card_id("AUTHOR")+1;
       COMPND_TOKEN = find_card_id("COMPND")+1;
       CRMETH_TOKEN = find_card_id("CRMETH")+1;
       CRSOLU_TOKEN = find_card_id("CRSOLU")+1;
       CRYST1_TOKEN = find_card_id("CRYST1")+1;
       DTMEAS_TOKEN = find_card_id("DTMEAS")+1;
       DTMETH_TOKEN = find_card_id("DTMETH")+1;
       DTWAVE_TOKEN = find_card_id("DTWAVE")+1;
       ECNUM_TOKEN  = find_card_id("ECNUM")+1;
       ENTAIL_TOKEN = find_card_id("ENTAIL")+1;
       ENTDES_TOKEN = find_card_id("ENTDES")+1;
       ENTITY_TOKEN = find_card_id("ENTITY")+1;
       ENTNAM_TOKEN = find_card_id("ENTNAM")+1;
       ENTMUT_TOKEN = find_card_id("ENTMUT")+1;
       EXPDTA_TOKEN = find_card_id("EXPDTA")+1;
       FORMUL_TOKEN = find_card_id("FORMUL")+1;
       FRAGMT_TOKEN = find_card_id("FRAGMT")+1;
       FREERF_TOKEN = find_card_id("FREERF")+1;
       HEADER_TOKEN = find_card_id("HEADER")+1;
       HELIX_TOKEN  = find_card_id("HELIX")+1;
       HET_TOKEN    = find_card_id("HET")+1;
       HETAC_TOKEN  = find_card_id("HETAC")+1;
       HETAD_TOKEN  = find_card_id("HETAD")+1;
       HETAI_TOKEN  = find_card_id("HETAI")+1;
       HETAIN_TOKEN = find_card_id("HETAIN")+1;
       HETAS_TOKEN  = find_card_id("HETAS")+1;
       HETIC_TOKEN  = find_card_id("HETIC")+1;
       HETNAM_TOKEN = find_card_id("HETNAM")+1;
       HETSYN_TOKEN = find_card_id("HETSYN")+1;
       HYDBND_TOKEN = find_card_id("HYDBND")+1;
       JRNL_TOKEN   = find_card_id("JRNL")+1;
       LINK_TOKEN   = find_card_id("LINK")+1;
       MASTER_TOKEN = find_card_id("MASTER")+1;
       MODRES_TOKEN = find_card_id("MODRES")+1;
       MTRIX_TOKEN  = find_card_id("MTRIX")+1;
       NDBFIL_TOKEN = find_card_id("NDBFIL")+1;
       NMRECD_TOKEN = find_card_id("NMRECD")+1;
       NMREXP_TOKEN = find_card_id("NMREXP")+1;
       NMRSFT_TOKEN = find_card_id("NMRSFT")+1;
       NMRSPM_TOKEN = find_card_id("NMRSPM")+1;
       ORIGX_TOKEN  = find_card_id("ORIGX")+1;
       PCOMPN_TOKEN = find_card_id("PCOMPN")+1;
       PDBFIL_TOKEN = find_card_id("PDBFIL")+1;
       PHVAL_TOKEN  = find_card_id("PHVAL")+1;
       PREMRK_TOKEN = find_card_id("PREMRK")+1;
       PROCES_TOKEN = find_card_id("PROCES")+1;
       PROTEI_TOKEN = find_card_id("PROTEI")+1;
       PSOURC_TOKEN = find_card_id("PSOURC")+1;
       RADIAT_TOKEN = find_card_id("RADIAT")+1;
       REFMET_TOKEN = find_card_id("REFMET")+1;
       REMARK_TOKEN = find_card_id("REMARK")+1;
       RENTRY_TOKEN = find_card_id("RENTRY")+1;
       REVDAT_TOKEN = find_card_id("REVDAT")+1;
       RFACTR_TOKEN = find_card_id("RFACTR")+1;
       SCALE_TOKEN  = find_card_id("SCALE")+1;
       SEQRES_TOKEN = find_card_id("SEQRES")+1;
       SFTWAR_TOKEN = find_card_id("SFTWAR")+1;
       SHEET_TOKEN  = find_card_id("SHEET")+1;
       SITE_TOKEN   = find_card_id("SITE")+1;
       SLTBRG_TOKEN = find_card_id("SLTBRG")+1;
       SOURCE_TOKEN = find_card_id("SOURCE")+1;
       SSBOND_TOKEN = find_card_id("SSBOND")+1;
       STATUS_TOKEN = find_card_id("STATUS")+1;
       STBIOL_TOKEN = find_card_id("STBIOL")+1;
       SYMTRA_TOKEN = find_card_id("SYMTRA")+1;
       TER_TOKEN    = find_card_id("TER")+1;
       TURN_TOKEN   = find_card_id("TURN")+1;
       WAVLEN_TOKEN = find_card_id("WAVLEN")+1;
}

static int read_cif_db()
{
       char binaryfile[300];
       sprintf(binaryfile, "%s/data/binary/%s\0", root_dir, cifbinfile);
       CifFileObj *fobjR = new CifFileObj(binaryfile, READ_MODE);
       if (!fobjR) return FALSE;

       if (!fobjR->IsTablePresent("NDB_CIF", "ndb_cif_category")) {
            delete fobjR; return FALSE;
       }
       ISTable *Table = fobjR->GetTablePtr("NDB_CIF", "ndb_cif_category");
       MxCategory = Table->GetNumRows();
       CifCategory = (CIF_CATEGORY *) calloc(MxCategory , sizeof(CIF_CATEGORY));
       int cate_index = Table->GetColumnIndex("category");
       int xyz_index = Table->GetColumnIndex("xyz");
       int sf_index = Table->GetColumnIndex("sf");
       int corr_index = Table->GetColumnIndex("correspondence");
       int depo_index = Table->GetColumnIndex("deposition");
       int xray_index = Table->GetColumnIndex("xray_or_nmr");
       int gen_index = Table->GetColumnIndex("general");
       int tran_index = Table->GetColumnIndex("transfer");
       int irow = 0;
       CifString string, string1;
       for (irow = 0; irow < MxCategory; irow++) {
            if (!Table->GetCell(string, cate_index, irow))
                 strcpy(CifCategory[irow].category, string.Text());
            else CifCategory[irow].category[0] = '\0';

            if (!Table->GetCell(string, xyz_index, irow))
                 CifCategory[irow].xyz = atoi(string.Text());
            else CifCategory[irow].xyz = 0;

            if (!Table->GetCell(string, sf_index, irow))
                 CifCategory[irow].sf = atoi(string.Text());
            else CifCategory[irow].sf = 0;

            if (!Table->GetCell(string, corr_index, irow))
                 CifCategory[irow].coordence = atoi(string.Text());
            else CifCategory[irow].coordence = 0;

            if (!Table->GetCell(string, depo_index, irow))
                 CifCategory[irow].deposition = atoi(string.Text());
            else CifCategory[irow].deposition = 0;

            if (!Table->GetCell(string, xray_index, irow))
                 CifCategory[irow].xray_or_nmr = atoi(string.Text());
            else CifCategory[irow].xray_or_nmr = 0;

            if (!Table->GetCell(string, gen_index, irow))
                 CifCategory[irow].general = atoi(string.Text());
            else CifCategory[irow].general = 0;

            if (!Table->GetCell(string, tran_index, irow))
                 CifCategory[irow].transfer = atoi(string.Text());
            else CifCategory[irow].transfer = 0;

            CifCategory[irow].CifKeyField = NULL;
            CifCategory[irow].nCifKeyField = 0;
            CifCategory[irow].keywords = NULL;
            CifCategory[irow].nkeywords = 0;
       }
       qsort((void *) CifCategory, MxCategory, sizeof(CIF_CATEGORY), ndb_category_compare);

       if (!fobjR->IsTablePresent("NDB_CIF", "ndb_cif_keyword")) {
            delete fobjR; return FALSE;
       }
       Table = fobjR->GetTablePtr("NDB_CIF", "ndb_cif_keyword");
       cate_index = Table->GetColumnIndex("category");
       int word_index = Table->GetColumnIndex("keyword");
       int token_index = Table->GetColumnIndex("ndb_token");
       int type_index = Table->GetColumnIndex("special_type");
       int field_index = Table->GetColumnIndex("ndb_field");
       int key_index = Table->GetColumnIndex("key");
       depo_index = Table->GetColumnIndex("deposition");

       CifString category("XXX", 4);
       CIF_CATEGORY *cifcategory = NULL;
       CifString keyword("XXX", 4);
       int row1 = 0, row2 = 0, l = 0, nfield, i, nrow;
       for (irow = 0; irow < Table->GetNumRows(); irow++) {
            Table->GetCell(string, cate_index, irow);
            if (string != category) {
                 category = string;
                 nfield = 0;
                 i = irow;
                 keyword.Clear();
                 keyword.Copy("XXX");
                 do {
                      Table->GetCell(string, cate_index, i);
                      if (string == category) {
                           Table->GetCell(string1, word_index, i);
                           if (string == category) {
                               if (string1 != keyword) {
                                    keyword = string1;
                                    nfield++;
                               }
                           }
                           i++;
                      }
                 } while (i < Table->GetNumRows() && string == category);
                 nrow = i;
                 cifcategory = find_category(category.Text());
                 if (cifcategory) {
                      cifcategory->nkeywords = nfield;
                      cifcategory->keywords = (CIF_KEYWORD *)
                                      calloc (nfield, sizeof(CIF_KEYWORD));
                      cifcategory->CifKeyField =
                                    (int *) calloc(nfield, sizeof(int));
                      for (i = 0; i < nfield; i++) {
                           cifcategory->CifKeyField[i] = 0;
                           cifcategory->keywords[i].numNdbInfo = 0;
                           cifcategory->keywords[i].name = NULL;
                           cifcategory->keywords[i].NdbInfo = NULL;
                      }
                      Table->GetCell(keyword, word_index, irow);
                      row1 = 0; nfield = 0;
                      for (i = irow; i < nrow; i++) {
                           Table->GetCell(string, word_index, i);
                           if (string != keyword) {
                                keyword = string;
                                cifcategory->keywords[row1].numNdbInfo = nfield;
                                cifcategory->keywords[row1].NdbInfo = (ndb_token_match *)
                                               calloc(nfield, sizeof(ndb_token_match));
                                nfield = 0;
                                row1++;
                           }
                           if (string == keyword) nfield++;
                      }
                      if (nfield) {
                           cifcategory->keywords[row1].numNdbInfo = nfield;
                           cifcategory->keywords[row1].NdbInfo =
                                 (ndb_token_match *) calloc(nfield, sizeof(ndb_token_match));
                      }
                 }
                 row1 = row2 = 0;
                 Table->GetCell(keyword, word_index, irow);
                 cifcategory->keywords[row1].name = (char *)
                            calloc(keyword.Length()+1, sizeof(char));
                 strcpy(cifcategory->keywords[row1].name, keyword.Text());
                 for (i = irow; i < nrow; i++) {
                      if (!Table->GetCell(string, word_index, i)) {
                           if (string != keyword) {
                                row1++;
                                keyword = string;
                                cifcategory->keywords[row1].name = (char *)
                                        calloc(keyword.Length()+1, sizeof(char));
                                strcpy(cifcategory->keywords[row1].name, keyword.Text());
                                row2 = 0;
                           }
                      }
              
                      cifcategory->keywords[row1].NdbInfo[row2].NdbTokenKeyFieldNo = NULL;
                      cifcategory->keywords[row1].NdbInfo[row2].nNdbTokenKeyFieldNo = 0;
     
                      if (!Table->GetCell(string, token_index, i)) {
                           cifcategory->keywords[row1].NdbInfo[row2].NdbTokenName = (char *)
                                    calloc(string.Length()+1, sizeof(char));
                           if (string == "?" || string == ".") {
                                cifcategory->keywords[row1].NdbInfo[row2].NdbTokenName[0] = '\0';
                                cifcategory->keywords[row1].NdbInfo[row2].NdbTokenId = -1;
                           } else {
                                strcpy(cifcategory->keywords[row1].NdbInfo[row2].NdbTokenName,
                                        string.Text());
                                cifcategory->keywords[row1].NdbInfo[row2].NdbTokenId =
                                        find_card_id(string.Text()) + 1;
                           }
                      } else {
                           cifcategory->keywords[row1].NdbInfo[row2].NdbTokenName = (char *)
                                 calloc(1, sizeof(char));
                           cifcategory->keywords[row1].NdbInfo[row2].NdbTokenName[0] = '\0';
                           cifcategory->keywords[row1].NdbInfo[row2].NdbTokenId = -1;
                      }
     
                      if (!Table->GetCell(string, type_index, i)) {
                           cifcategory->keywords[row1].NdbInfo[row2].JrnlTokenName_Type
                                = (char *) calloc(string.Length()+1, sizeof(char));
                           if (string == "?" || string == ".") {
                                cifcategory->keywords[row1].NdbInfo[row2].JrnlTokenName_Type[0] = '\0';
                                cifcategory->keywords[row1].NdbInfo[row2].JrnlTokenId = -1;
                           } else {
                                strcpy(cifcategory->keywords[row1].NdbInfo[row2].JrnlTokenName_Type,
                                        string.Text());
                                if (cifcategory->keywords[row1].NdbInfo[row2].NdbTokenId == JRNL_TOKEN)
                                     cifcategory->keywords[row1].NdbInfo[row2].JrnlTokenId =
                                        find_jrnl_id(string.Text()) + 1;
                                else cifcategory->keywords[row1].NdbInfo[row2].JrnlTokenId = - 1;
                           }
                      } else {
                           cifcategory->keywords[row1].NdbInfo[row2].JrnlTokenName_Type
                                = (char *) calloc(string.Length()+1, sizeof(char));
                           cifcategory->keywords[row1].NdbInfo[row2].JrnlTokenName_Type[0] = '\0';
                           cifcategory->keywords[row1].NdbInfo[row2].JrnlTokenId = -1;
                      }
     
                      if (!Table->GetCell(string, field_index, i)) {
                           if (string == "?" || string == ".")
                                cifcategory->keywords[row1].NdbInfo[row2].NdbFieldNo = -1;
                           else cifcategory->keywords[row1].NdbInfo[row2].NdbFieldNo =
                                      atoi(string.Text());
                      } else cifcategory->keywords[row1].NdbInfo[row2].NdbFieldNo = -1;
     
                      if (!Table->GetCell(string, key_index, i)) {
                           cifcategory->keywords[row1].NdbInfo[row2].IsKey = atoi(string.Text());
                           if (cifcategory->keywords[row1].NdbInfo[row2].IsKey != 1)
                                cifcategory->keywords[row1].NdbInfo[row2].IsKey = 0;
                           if (cifcategory->keywords[row1].NdbInfo[row2].IsKey &&
                               keyword != "entry_id") {
                                for (l = 0; l < cifcategory->nCifKeyField; l++)
                                     if (row1 == cifcategory->CifKeyField[l]) break;
                                if (l == cifcategory->nCifKeyField) {
                                     cifcategory->CifKeyField[l] = row1;
                                     cifcategory->nCifKeyField++;
                                }
                           }
                      } else cifcategory->keywords[row1].NdbInfo[row2].IsKey = 0;
     
                      if (!Table->GetCell(string, depo_index, i))
                           cifcategory->keywords[row1].NdbInfo[row2].deposition = atoi(string.Text());
                      else cifcategory->keywords[row1].NdbInfo[row2].deposition = 0;
     
                      row2++;
                 }
                 irow = nrow - 1;
            }
       }
       delete fobjR;

       int Id = 0, j = 0, k = 0, keyField, keyTokenId, KeyFieldNo, TokenId;
       for (Id = 0; Id < MxCategory; Id++) {
            if (CifCategory[Id].nCifKeyField) {
                 for (i = 0; i < CifCategory[Id].nkeywords; i++) {
                      for (j = 0; j < CifCategory[Id].keywords[i].numNdbInfo; j++) {
                           CifCategory[Id].keywords[i].NdbInfo[j].NdbTokenKeyFieldNo
                            = (int *) calloc(CifCategory[Id].nCifKeyField, sizeof(int));
                           for (l = 0; l < CifCategory[Id].nCifKeyField; l++)
                                CifCategory[Id].keywords[i].NdbInfo[j].
                                            NdbTokenKeyFieldNo[l] = 0;
     
                           for (l = 0; l < CifCategory[Id].nCifKeyField; l++) {
                                keyField = CifCategory[Id].CifKeyField[l];
                                keyTokenId = -1;
                                KeyFieldNo = -1;
                                for (k = 0; k < CifCategory[Id].keywords[keyField].
                                           numNdbInfo; k++) {
                                     if (!strcmp(CifCategory[Id].keywords[keyField].
                                                 NdbInfo[k].JrnlTokenName_Type,
                                                 CifCategory[Id].keywords[i].
                                                 NdbInfo[j].JrnlTokenName_Type)) break;
                                }
                                if (k < CifCategory[Id].keywords[keyField].numNdbInfo) {
                                     keyTokenId = CifCategory[Id].keywords[keyField].
                                                      NdbInfo[k].NdbTokenId - 1;
                                     KeyFieldNo = CifCategory[Id].keywords[keyField].
                                                      NdbInfo[k].NdbFieldNo - 1;
                                }
                                TokenId = CifCategory[Id].keywords[i].NdbInfo[j].
                                              NdbTokenId - 1;
                                if (TokenId == JRNL_TOKEN - 1) {
                                     TokenId = CifCategory[Id].keywords[i].NdbInfo[0].
                                                    JrnlTokenId - 1;
                                     keyTokenId = CifCategory[Id].keywords[keyField].
                                                    NdbInfo[0].JrnlTokenId - 1;
                                     KeyFieldNo = CifCategory[Id].keywords[keyField].
                                                    NdbInfo[0].NdbFieldNo - 1;
                                     if (TokenId <0 || keyTokenId < 0 || KeyFieldNo < 0) continue;
                                     for (k = 0; k < JrnlToken[TokenId].NumField; k++) {
                                          if (!strcmp(JrnlToken[TokenId].FieldList[k].FieldLabel,
                                              JrnlToken[keyTokenId].FieldList[KeyFieldNo].FieldLabel))
                                               break;
                                     }
                                     if (k < JrnlToken[TokenId].NumField)
                                          CifCategory[Id].keywords[i].NdbInfo[j].
                                             NdbTokenKeyFieldNo[l] = k;
                                } else if (keyTokenId >= 0 && KeyFieldNo >= 0 && 
                                    TokenId != JRNL_TOKEN - 1 && TokenId >= 0) {
                                     for (k = 0; k < NdbToken[TokenId].NumField; k++) {
                                          if (!strcmp(NdbToken[TokenId].FieldList[k].FieldLabel,
                                              NdbToken[keyTokenId].FieldList[KeyFieldNo].FieldLabel))
                                               break;
                                     }
                                     if (k < NdbToken[TokenId].NumField)
                                          CifCategory[Id].keywords[i].NdbInfo[j].
                                             NdbTokenKeyFieldNo[l] = k;
                                }
                           }
                      }
                 }
            }
       }
       return TRUE;
}	

static int read_cif_rms()
{
       char binaryfile[300];
       sprintf(binaryfile, "%s/data/binary/%s\0", root_dir, rmsbinfile);
       CifFileObj *fobjR = new CifFileObj(binaryfile, READ_MODE);
       if (!fobjR) return FALSE;

       if (!fobjR->IsTablePresent("NDB_CIF_RMS", "ndb_cif_program")) {
            delete fobjR; return FALSE;
       }
       ISTable *Table = fobjR->GetTablePtr("NDB_CIF_RMS", "ndb_cif_program");
       int full_index = Table->GetColumnIndex("full_name");
       int abbr_index = Table->GetColumnIndex("abbreviation");
       mxPrograms = Table->GetNumRows();
       programs = (CIF_PROGRAMS *) calloc(mxPrograms, sizeof(CIF_PROGRAMS));
       int irow = 0, j;
       CifString string;
       for (irow = 0; irow < mxPrograms; irow++) {
            if (!Table->GetCell(string, full_index, irow)) {
                 if (string == "." || string == "?")
                      programs[irow].program[0] = '\0';
                 else strcpy(programs[irow].program, string.Text());
            }
            if (!Table->GetCell(string, abbr_index, irow)) {
                 if (string == "." || string == "?")
                      programs[irow].abbreviation[0] = '\0';
                 else strcpy(programs[irow].abbreviation, string.Text());
            }
       }

       if (!fobjR->IsTablePresent("NDB_CIF_RMS", "ndb_cif_rms")) {
            delete fobjR; return FALSE;
       }
       Table = fobjR->GetTablePtr("NDB_CIF_RMS", "ndb_cif_rms");
       mxRmsRecords = Table->GetNumRows();
       rmsRecords = (NDBCIF_RMS_RECORDS *) calloc(mxRmsRecords,
                               sizeof(NDBCIF_RMS_RECORDS));
       for (irow = 0; irow < mxRmsRecords; irow++) {
            Table->GetCell(string, 0, irow);
            strcpy(rmsRecords[irow].type, string.Text());
            for (j = 0; j < NDB_CIF_RMS_ITEMS; j++) {
                 Table->GetCell(string, 2 * j + 1, irow);
                 rmsRecords[irow].tokens[j] = find_card_id(string.Text());
                 Table->GetCell(string, 2 * j + 2, irow);
                 if (string == "." || string == "?")
                      rmsRecords[irow].fields[j] = -1;
                 else rmsRecords[irow].fields[j] = atoi(string.Text()) - 1;
            }
       }
       if (fobjR) delete fobjR;

       return TRUE;
}

static int read_refinement_program()
{
       char binaryfile[300];
       sprintf(binaryfile, "%s/data/binary/%s\0", root_dir, refinementbinfile);
       CifFileObj *fobjR = new CifFileObj(binaryfile, READ_MODE);
       if (!fobjR) return FALSE;

       if (!fobjR->IsTablePresent("NDB_REFINEMENT", "refinement")) {
            delete fobjR; return FALSE;
       }
       ISTable *Table = fobjR->GetTablePtr("NDB_REFINEMENT", "refinement");
       Max_Refinement_Programs = Table->GetNumRows();
       Refinement_programs = (REFMET_AUTHOR *) calloc(Max_Refinement_Programs,
                   sizeof(REFMET_AUTHOR));
       int prog_index = Table->GetColumnIndex("program");
       int auth_index = Table->GetColumnIndex("author");
       int irow;
       CifString string;
       for (irow = 0; irow < Max_Refinement_Programs; irow++) {
            Table->GetCell(string, prog_index, irow);
            strcpy(Refinement_programs[irow].Program, string.Text());
            Table->GetCell(string, auth_index, irow);
            strcpy(Refinement_programs[irow].Author, string.Text());
            ndb_clean_string(Refinement_programs[irow].Program);
            ndb_clean_string(Refinement_programs[irow].Author);
       }
       if (fobjR) delete fobjR;

       return TRUE;
}

static int ndb_category_compare(const void *a, const void *b)
{
       return (strcmp(((CIF_CATEGORY *) a)->category,
                      ((CIF_CATEGORY *) b)->category));
}

