/*
FILE:     fileobj.C
*/
/*
VERSION:  0.904-beta
*/
/*
DATE:     8/12/2003
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2003 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>

#include "ndb_misclib.h"
#include "fileobj.h"
#include "util.h"

_FileObj::_FileObj(const int mxtokens, const int mxjrnltokens)
{
       _input_filename[0] = '\0';
       _output_filename[0] = '\0';
       _temp_file[0] = '\0';
       _StructureId[0] = '\0';
       _experiment_type = EXPERIMENT_TYPE_BASIC;
       _num_mol = 0;
       _alloc_mol = 0;
       _CardInfo = new Cards_Info_Format[mxtokens];
       _JrnlCardInfo = new Cards_Info_Format[mxjrnltokens];
       _molecules = NULL;
       _CifObj = NULL;
       _Alloc_Seqs = 0;
       _Num_Seqs = 0;
       _Seqs = NULL;
       _cell = NULL;
       _error_message.Copy("");
       _alloc_link = 0; _n_link = 0; _link = NULL;
       _alloc_sheet = 0; _n_sheet = 0; _sheet = NULL;
       _alloc_ssbond = 0; _n_ssbond = 0; _ssbond = NULL;
       _alloc_hydbnd = 0; _n_hydbnd = 0; _hydbnd = NULL;
       _alloc_sltbrg = 0; _n_sltbrg = 0; _sltbrg = NULL;
       _alloc_site = 0; _n_site = 0; _site = NULL;
       _alloc_het = 0; _n_het = 0; _het = NULL;
       _alloc_modres = 0; _n_modres = 0; _modres = NULL;
}

_FileObj::~_FileObj()
{
       int i, j;
       if (_error_message != "") {
            ndb_log_message(NDB_MSG_ERR, "%s\n", _error_message.Text());
       }
       delete [] _CardInfo;
       delete [] _JrnlCardInfo;
       if (_molecules) {
            for (i = 0; i < _num_mol; i++)
                 if (_molecules[i]) delete _molecules[i];
            delete [] _molecules;
       }
       if(_CifObj) { _CifObj->CloseFile(); delete _CifObj; }
       if (_Seqs) {
            for (i = 0; i < _Num_Seqs; i++) {
                 if (_Seqs[i]) {
                      for (j = 0; j < _Seqs[i]->SeqLen; j++)
                           if (_Seqs[i]->res[j]) delete _Seqs[i]->res[j];
                      delete [] _Seqs[i]->res;
                      delete _Seqs[i];
                 }
            }
            delete [] _Seqs;
       }
       if (_cell) {
            if (_cell->symops) delete [] _cell->symops;
            delete _cell;
       }
       if (_link) {
            for (i = 0; i < _n_link; i++) {
                 if (_link[i]) {
                      if (_link[i]->details) delete [] _link[i]->details;
                      delete _link[i];
                 }
            }
            delete [] _link;
       }
       if (_sheet) {
            for (i = 0; i < _n_sheet; i++)
                 if (_sheet[i]) delete _sheet[i];
            delete [] _sheet;
       }
       if (_ssbond) {
            for (i = 0; i < _n_ssbond; i++)
                 if (_ssbond[i]) delete _ssbond[i];
            delete [] _ssbond;
       }
       if (_hydbnd) {
            for (i = 0; i < _n_hydbnd; i++)
                 if (_hydbnd[i]) delete _hydbnd[i];
            delete [] _hydbnd;
       }
       if (_sltbrg) {
            for (i = 0; i < _n_sltbrg; i++) {
                 if (_sltbrg[i]) {
                      if (_sltbrg[i]->details) delete _sltbrg[i]->details;
                      delete _sltbrg[i];
                 }
            }
            delete [] _sltbrg;
       }
       if (_site) {
            for (i = 0; i < _n_site; i++) {
                 if (_site[i]) {
                      if (_site[i]->initIndex)
                           delete [] _site[i]->initIndex;
                      if (_site[i]->initRes)
                           delete [] _site[i]->initRes;
                      delete _site[i];
                 }
            }
            delete [] _site;
       }
       if (_het) {
            for (i = 0; i < _n_het; i++)
                 if (_het[i]) delete _het[i];
            delete [] _het;
       }
       if (_modres) {
            for (i = 0; i < _n_modres; i++) {
                 if (_modres[i]) {
                      if (_modres[i]->details) delete [] _modres[i]->details;
                      delete _modres[i];
                 }
            }
            delete [] _modres;
       }
}

void _FileObj::_ManageSpace()
{
       int i = 0;
       if (_num_mol >= _alloc_mol) {
            _alloc_mol += 10;
            _MOLECULE **tmp = _molecules;
            _molecules = new _MOLECULE*[_alloc_mol];
            for (i = 0; i < _num_mol; i++) _molecules[i] = tmp[i];
            for (i = _num_mol; i < _alloc_mol; i++) _molecules[i] = NULL;
            delete [] tmp;
       }
       _molecules[_num_mol] = new _MOLECULE;
       _num_mol++;
}

void _FileObj::_ManageSeqSpace()
{
       int i = 0;
       if (_Num_Seqs >= _Alloc_Seqs) {
            _Alloc_Seqs += 10;
            SEQ** tmp = _Seqs;
            _Seqs = new SEQ*[_Alloc_Seqs];
            for (i = 0; i < _Num_Seqs; i++) _Seqs[i] = tmp[i];
            for (i = _Num_Seqs; i < _Alloc_Seqs; i++) _Seqs[i] = NULL;
            delete [] tmp;
       }
       _Seqs[_Num_Seqs] = new SEQ;
       memset(_Seqs[_Num_Seqs]->ChainId, 0, 5);
       _Seqs[_Num_Seqs]->AllocLen = 0;
       _Seqs[_Num_Seqs]->SeqLen = 0;
       _Seqs[_Num_Seqs]->chain_type = -1;
       _Seqs[_Num_Seqs]->res = NULL;
       _Num_Seqs++;
}

void _FileObj::_ManageResSpace(SEQ *seq)
{
       int i = 0;
       if (seq->SeqLen >= seq->AllocLen) {
            seq->AllocLen += 100;
            RES **tmp = seq->res;
            seq->res = new RES*[seq->AllocLen];
            for (i = 0; i < seq->SeqLen; i++) seq->res[i] = tmp[i];
            for (i = seq->SeqLen; i < seq->AllocLen; i++) seq->res[i] = NULL;
            delete [] tmp;
       }
}

void  _FileObj::set_input_filename(const char *string)  { strcpy(_input_filename, string); }
void  _FileObj::set_output_filename(const char *string) { strcpy(_output_filename, string); }

void _FileObj::get_sequence_types()
{
       int type = 0, types[4], i = 0, j = 0, max_num = -1, index = -1;
       for (i = 0; i < _Num_Seqs; i++) {
            for (j = 0; j < 4; j++) types[j] = 0;
            for (j = 0; j < _Seqs[i]->SeqLen; j++) {
                 type = find_residue_type(_Seqs[i]->res[j]->name);
                 if (type == ATOMN_TOKEN) types[0]++;
                 else if (type == ATOMP_TOKEN) types[1]++;
                 else if (type == ATOMS_TOKEN) types[2]++;
                 else if (type == HETAIN_TOKEN) types[3]++;
            }
            max_num = -1;
            index = -1;
            for (j = 0; j < 4; j++) {
                 if (types[j] > max_num) {
                      max_num = types[j];
                      index = j;
                 }
            }
            if (index >= 0) {
                 if (index == 0) _Seqs[i]->chain_type = ATOMN_TOKEN;
                 else if (index == 1) _Seqs[i]->chain_type = ATOMP_TOKEN;
                 else if (index == 2) _Seqs[i]->chain_type = ATOMS_TOKEN;
                 else if (index == 3) _Seqs[i]->chain_type = HETAIN_TOKEN;
            }
       }
}

void _FileObj::get_seqres_from_seqs(int& chain_flag)
{
       int i = 0, j = 0, seqField = 4, seqNo = 1;
       _CHAIN *chain = NULL;
       One_Card_Info_Format *pCard = NULL;

       // if (chain_flag == 1 && _Num_Seqs > 1) chain_flag = 0;
       if (chain_flag == 1) {
            chain_flag = 0;
            chain = _molecules[0]->GetFirstChain();
            while (chain) {
                 if ((chain->chain_type() == ATOMP_TOKEN ||
                      chain->chain_type() == ATOMN_TOKEN) &&
                     (!strcmp(chain->PDB_ChainID_Flag(), "Y") ||
                      !strcmp(chain->PDB_ChainID_Flag(), "y"))) {
                      chain_flag = 1;
                      break;
                 }
                 chain = _molecules[0]->GetNextChain();
            }
       }

       for (i = 0; i < _Num_Seqs; i++) {
            chain = _molecules[0]->GetFirstChain();
            while (chain) {
                 if (!strcmp(chain->ChainID(), _Seqs[i]->ChainId)) {
                      if (chain_flag == 2)
                           strcpy(_Seqs[i]->ChainId, chain->PUB_ChainID());
                      else if (chain_flag == 1)
                           strcpy(_Seqs[i]->ChainId, "");
                      else strcpy(_Seqs[i]->ChainId, chain->PDB_ChainID());
                      break;
                 }
                 chain = _molecules[0]->GetNextChain();
            }
            seqNo = 1;
            pCard = _CardInfo[SEQRES_TOKEN-1].
                      add_a_new_card(NdbToken[SEQRES_TOKEN-1], 0);
            pCard->AddElementToField(seqNo, 1);
            pCard->AddElementToField(_Seqs[i]->ChainId, 2);
            pCard->AddElementToField(_Seqs[i]->SeqLen, 3);
            seqField = 4;
            for (j = 0; j < _Seqs[i]->SeqLen; j++) {
                 if (seqField == NdbToken[SEQRES_TOKEN-1].NumField) {
                      pCard = _CardInfo[SEQRES_TOKEN-1].
                                add_a_new_card(NdbToken[SEQRES_TOKEN-1], 0);
                      seqNo++;
                      pCard->AddElementToField(seqNo, 1);
                      pCard->AddElementToField(_Seqs[i]->ChainId, 2);
                      pCard->AddElementToField(_Seqs[i]->SeqLen, 3);
                      seqField = 4;
                 }
                 pCard->AddElementToField(_Seqs[i]->res[j]->name, seqField);
                 seqField++;
            }
       }
}

void _FileObj::delete_null_cards()
{
       int i = 0, j = 0;
       One_Card_Info_Format *pCard = NULL;
       for (i = 0; i < MxTokens; i++) {
            if (i == REMARK_TOKEN-1 || i == PREMRK_TOKEN-1 ||
                i == PCOMPN_TOKEN-1 || i == PSOURC_TOKEN-1) continue;

            pCard = _CardInfo[i].Head();
            while (pCard) {
                 for (j = 1; j < NdbToken[i].NumField; j++) {
                      if (j != NdbToken[i].SeqField-1 && j != NdbToken[i].ConField-1 &&
                          strcmp(pCard->Field(j), "") && (j != 2 || (i != FORMUL_TOKEN-1 &&
                          i != HETSYN_TOKEN-1 && i != HETNAM_TOKEN-1))) break;
                 }
                 if (j == NdbToken[i].NumField)
                      pCard = _CardInfo[i].delete_a_card(pCard);
                 else pCard = pCard->Next();
            }
       }
       for (i = 0; i < MxJrnlTokens; i++) {
            pCard = _JrnlCardInfo[i].Head();
            while (pCard) {
                 for (j = 4; j < JrnlToken[i].NumField; j++) {
                      if (strcmp(pCard->Field(j), "")) break;
                 }
                 if (j == JrnlToken[i].NumField)
                      pCard = _JrnlCardInfo[i].delete_a_card(pCard);
                 else pCard = pCard->Next();
            }
       }
}

void _FileObj::update_tokens()
{
       int card_id = 0, MatrixNo = 0, SerialNo = 0, field_no = 0, width = 0;
       char new_date[20];
       One_Card_Info_Format *pCard = NULL;

       for (card_id = 0; card_id < MxTokens; card_id++) {
            MatrixNo = 1;
            SerialNo = 0;
            pCard = _CardInfo[card_id].Head();
            if (NdbToken[card_id].MatrixType == 0 && NdbToken[card_id].SeqField != 0) {
                 while (pCard) {
                      SerialNo++;
                      if (!strcmp(pCard->Field(NdbToken[card_id].SeqField-1), "")) {
                           pCard->AddElementToField(SerialNo, NdbToken[card_id].SeqField-1);
                      }
                      pCard = pCard->Next();
                 }
            } else if (NdbToken[card_id].MatrixType) {
                 while (pCard) {
                      sprintf(new_date, "%s%d\0", pCard->Field(0), MatrixNo);
                      pCard->AddElementToField(new_date, 0);
                      pCard = pCard->Next();
                      if (MatrixNo % 3 == 0) MatrixNo = 1;
                      else MatrixNo++;
                 }
            }
            pCard = _CardInfo[card_id].Head();
            while (pCard) {
                 for (field_no = 1; field_no < NdbToken[card_id].NumField; field_no++) {
                      if (NdbToken[card_id].FieldList[field_no].FieldType == NDB_DATE_FIELD &&
                          strcmp(pCard->Field(field_no), "")) {
                           width = NdbToken[card_id].FieldList[field_no].FieldWidth;
                           if (card_id == DTMEAS_TOKEN-1) width++;
                           ndb_convert_time_format(pCard->Field(field_no), new_date, width);
                           pCard->AddElementToField(new_date, field_no);
                      }
                 }
                 pCard = pCard->Next();
            }
       }
}

void _FileObj::get_cell(void)
{
       if (_cell) return;

       _cell = new NDBCELLCONST;
       _cell->a = _cell->b = _cell->c = _cell->alpha = _cell->beta = _cell->gamma = 0.0;
       _cell->space_group[0] = '\0';
       _cell->full_space_group[0] = '\0';
       _cell->z = 0;
       _cell->is_artifical = 0;

       int exist_scale_matrix = 0;
       double scale[3][3], vec[3];

       int numBlock = 0;
       ISTable *Table = NULL;
       CifString string;
       char **blockName = _CifObj->GetBlockNames(numBlock);
       if (_CifObj->IsTablePresent(blockName[0], "symmetry")) {
            Table = _CifObj->GetTablePtr(blockName[0], "symmetry");
            int name_index = Table->GetColumnIndex("space_group_name_H-M");
            int full_name = Table->GetColumnIndex("ndb_full_space_group_name_H-M");
            string.Copy("");
            Table->GetCell(string, name_index, 0);
            if (string != "" && string != "." && string != "?")
                 strcpy(_cell->space_group, string.Text());
            string.Copy("");
            Table->GetCell(string, full_name, 0);
            if (string != "" && string != "." && string != "?")
                 strcpy(_cell->full_space_group, string.Text());
       }
       if (_CifObj->IsTablePresent(blockName[0], "cell")) {
            Table = _CifObj->GetTablePtr(blockName[0], "cell");
            int a_index = Table->GetColumnIndex("length_a");
            int b_index = Table->GetColumnIndex("length_b");
            int c_index = Table->GetColumnIndex("length_c");
            int alpha_index = Table->GetColumnIndex("angle_alpha");
            int beta_index = Table->GetColumnIndex("angle_beta");
            int gamma_index = Table->GetColumnIndex("angle_gamma");
            int z_index = Table->GetColumnIndex("Z_PDB");
            string.Copy("");
            Table->GetCell(string, a_index, 0);
            if (string != "" && string != "." && string != "?")
                 _cell->a = atof(string.Text());
            string.Copy("");
            Table->GetCell(string, b_index, 0);
            if (string != "" && string != "." && string != "?")
                 _cell->b = atof(string.Text());
            string.Copy("");
            Table->GetCell(string, c_index, 0);
            if (string != "" && string != "." && string != "?")
                 _cell->c = atof(string.Text());
            string.Copy("");
            Table->GetCell(string, alpha_index, 0);
            if (string != "" && string != "." && string != "?")
                 _cell->alpha = atof(string.Text());
            string.Copy("");
            Table->GetCell(string, beta_index, 0);
            if (string != "" && string != "." && string != "?")
                 _cell->beta = atof(string.Text());
            string.Copy("");
            Table->GetCell(string, gamma_index, 0);
            if (string != "" && string != "." && string != "?")
                 _cell->gamma = atof(string.Text());
            string.Copy("");
            Table->GetCell(string, z_index, 0);
            if (string != "" && string != "." && string != "?")
                 _cell->z = atoi(string.Text());
       }
       if (_CifObj->IsTablePresent(blockName[0], "atom_sites")) {
            Table = _CifObj->GetTablePtr(blockName[0], "atom_sites");
            int index_11 = Table->GetColumnIndex("fract_transf_matrix[1][1]");
            int index_12 = Table->GetColumnIndex("fract_transf_matrix[1][2]");
            int index_13 = Table->GetColumnIndex("fract_transf_matrix[1][3]");
            int index_21 = Table->GetColumnIndex("fract_transf_matrix[2][1]");
            int index_22 = Table->GetColumnIndex("fract_transf_matrix[2][2]");
            int index_23 = Table->GetColumnIndex("fract_transf_matrix[2][3]");
            int index_31 = Table->GetColumnIndex("fract_transf_matrix[3][1]");
            int index_32 = Table->GetColumnIndex("fract_transf_matrix[3][2]");
            int index_33 = Table->GetColumnIndex("fract_transf_matrix[3][3]");
            int index_v1 = Table->GetColumnIndex("fract_transf_vector[1]");
            int index_v2 = Table->GetColumnIndex("fract_transf_vector[2]");
            int index_v3 = Table->GetColumnIndex("fract_transf_vector[3]");
            exist_scale_matrix = 1;
            string.Copy("");
            Table->GetCell(string, index_11, 0);
            if (string != "" && string != "." && string != "?")
                 scale[0][0] = atof(string.Text());
            else exist_scale_matrix = 0;

            string.Copy("");
            Table->GetCell(string, index_12, 0);
            if (string != "" && string != "." && string != "?")
                 scale[0][1] = atof(string.Text());
            else exist_scale_matrix = 0;

            string.Copy("");
            Table->GetCell(string, index_13, 0);
            if (string != "" && string != "." && string != "?")
                 scale[0][2] = atof(string.Text());
            else exist_scale_matrix = 0;

            string.Copy("");
            Table->GetCell(string, index_21, 0);
            if (string != "" && string != "." && string != "?")
                 scale[1][0] = atof(string.Text());
            else exist_scale_matrix = 0;

            string.Copy("");
            Table->GetCell(string, index_22, 0);
            if (string != "" && string != "." && string != "?")
                 scale[1][1] = atof(string.Text());
            else exist_scale_matrix = 0;

            string.Copy("");
            Table->GetCell(string, index_23, 0);
            if (string != "" && string != "." && string != "?")
                 scale[1][2] = atof(string.Text());
            else exist_scale_matrix = 0;

            string.Copy("");
            Table->GetCell(string, index_31, 0);
            if (string != "" && string != "." && string != "?")
                 scale[2][0] = atof(string.Text());
            else exist_scale_matrix = 0;

            string.Copy("");
            Table->GetCell(string, index_32, 0);
            if (string != "" && string != "." && string != "?")
                 scale[2][1] = atof(string.Text());
            else exist_scale_matrix = 0;

            string.Copy("");
            Table->GetCell(string, index_33, 0);
            if (string != "" && string != "." && string != "?")
                 scale[2][2] = atof(string.Text());
            else exist_scale_matrix = 0;

            string.Copy("");
            Table->GetCell(string, index_v1, 0);
            if (string != "" && string != "." && string != "?")
                 vec[0] = atof(string.Text());
            else exist_scale_matrix = 0;

            string.Copy("");
            Table->GetCell(string, index_v2, 0);
            if (string != "" && string != "." && string != "?")
                 vec[1] = atof(string.Text());
            else exist_scale_matrix = 0;

            string.Copy("");
            Table->GetCell(string, index_v3, 0);
            if (string != "" && string != "." && string != "?")
                 vec[2] = atof(string.Text());
            else exist_scale_matrix = 0;
       }
       for (int i = 0; i < numBlock; i++) delete [] blockName[i];
       delete [] blockName;

       if (_cell->a == 1.0 && _cell->b == 1.0 && _cell->c == 1.0 &&
           _cell->alpha == 90.0 && _cell->beta == 90.0 && _cell->gamma == 90.0 &&
           !strcmp(_cell->space_group, "P 1")) {
            _cell->a = 1000.0;
            _cell->b = 1000.0;
            _cell->c = 1000.0;
            _cell->is_artifical = 1;
       } else if (_cell->a == 0.0 || _cell->b == 0.0 || _cell->c == 0.0 ||
                  _cell->alpha == 0.0 || _cell->beta == 0.0 || _cell->gamma == 0.0 ||
                  !strcmp(_cell->space_group, "")) {
            _cell->a = 1000.0;
            _cell->b = 1000.0;
            _cell->c = 1000.0;
            _cell->alpha = 90.0;
            _cell->beta = 90.0;
            _cell->gamma = 90.0;
            strcpy(_cell->space_group, "P 1");
            _cell->is_artifical = 1;
       }
       ndb_get_sym_data(_cell);
       ndb_ortho_to_frac_matrix(_cell);
       ndb_frac_to_ortho_matrix(_cell);

       if (!_cell->is_artifical && exist_scale_matrix)
            check_scale_token(_cell, scale, vec);
}

void _FileObj::Insert_a_Sheet(const char *id, const int numStrands, const int initIndex,
               const int initRes, const int endIndex, const int endRes, const int sense,
               const char *curAtom, const int curIndex, const int curRes,
               const char *prevAtom, const int prevIndex, const int prevRes)
{
       int i = 0;
       if (_n_sheet >= _alloc_sheet) {
            _alloc_sheet += 20;
            _SHEET** tmp = _sheet;
            _sheet = new _SHEET*[_alloc_sheet];
            for (i = 0; i < _n_sheet; i++) _sheet[i] = tmp[i];
            for (i = _n_sheet; i < _alloc_sheet; i++) _sheet[i] = NULL;
            delete [] tmp;
       }
       _sheet[_n_sheet] = new _SHEET;
       strcpy(_sheet[_n_sheet]->sheetID, id);
       _sheet[_n_sheet]->numStrands = numStrands;
       _sheet[_n_sheet]->initIndex = initIndex;
       _sheet[_n_sheet]->initRes = initRes;
       _sheet[_n_sheet]->endIndex = endIndex;
       _sheet[_n_sheet]->endRes = endRes;
       _sheet[_n_sheet]->sense = sense;
       strcpy(_sheet[_n_sheet]->curAtom, curAtom);
       _sheet[_n_sheet]->curIndex = curIndex;
       _sheet[_n_sheet]->curRes = curRes;
       strcpy(_sheet[_n_sheet]->prevAtom, prevAtom);
       _sheet[_n_sheet]->prevIndex = prevIndex;
       _sheet[_n_sheet]->prevRes = prevRes;
       _n_sheet++;
}

void _FileObj::Insert_a_SSbond(const int FstIndex, const int FstRes, const int SndIndex,
                  const int SndRes, const char *SymOP_1, const char *SymOP_2)
{
       int i = 0;
       for (i = 0; i < _n_ssbond; i++) {
            if (_ssbond[i]->FstIndex == FstIndex && _ssbond[i]->FstRes == FstRes &&
                _ssbond[i]->SndIndex == SndIndex && _ssbond[i]->SndRes == SndRes &&
                !strcmp(_ssbond[i]->SymOP_1, SymOP_1) &&
                !strcmp(_ssbond[i]->SymOP_2, SymOP_2)) break;
       }
       if (i < _n_ssbond) return;

       if (_n_ssbond >= _alloc_ssbond) {
            _alloc_ssbond += 5;
            _SSBOND** tmp = _ssbond;
            _ssbond = new _SSBOND*[_alloc_ssbond];
            for (i = 0; i < _n_ssbond; i++) _ssbond[i] = tmp[i];
            for (i = _n_ssbond; i < _alloc_ssbond; i++) _ssbond[i] = NULL;
            delete [] tmp;
       }
       _ssbond[_n_ssbond] = new _SSBOND;
       _ssbond[_n_ssbond]->FstIndex = FstIndex;
       _ssbond[_n_ssbond]->FstRes = FstRes;
       _ssbond[_n_ssbond]->SndIndex = SndIndex;
       _ssbond[_n_ssbond]->SndRes = SndRes;
       strcpy(_ssbond[_n_ssbond]->SymOP_1, SymOP_1);
       strcpy(_ssbond[_n_ssbond]->SymOP_2, SymOP_2);
       _n_ssbond++;
}

void _FileObj::Insert_a_Hydbnd(const char *FstAtom, const char *altLoc1, const int FstIndex,
           const int FstRes, const char *H_Atom, const char *altLocH, const int H_Index,
           const int H_Res, const char *SndAtom, const char *altLoc2, const int SndIndex,
           const int SndRes, const char *SymOP_1, const char *SymOP_2)
{
       int i = 0;
       for (i = 0; i < _n_hydbnd; i++) {
            if (!strcmp(_hydbnd[i]->FstAtom, FstAtom) &&
                !strcmp(_hydbnd[i]->altLoc1, altLoc1) &&
                _hydbnd[i]->FstIndex == FstIndex && _hydbnd[i]->FstRes == FstRes &&
                !strcmp(_hydbnd[i]->H_Atom, H_Atom) &&
                !strcmp(_hydbnd[i]->altLocH, altLocH) &&
                _hydbnd[i]->H_Index == H_Index && _hydbnd[i]->H_Res == H_Res &&
                !strcmp(_hydbnd[i]->SndAtom, SndAtom) &&
                !strcmp(_hydbnd[i]->altLoc2, altLoc2) &&
                _hydbnd[i]->SndIndex == SndIndex && _hydbnd[i]->SndRes == SndRes &&
                !strcmp(_hydbnd[i]->SymOP_1, SymOP_1) &&
                !strcmp(_hydbnd[i]->SymOP_2, SymOP_2)) break;
       }
       if (i < _n_hydbnd) return;

       if (_n_hydbnd >= _alloc_hydbnd) {
            _alloc_hydbnd += 5;
            _HYDBND** tmp = _hydbnd;
            _hydbnd = new _HYDBND*[_alloc_hydbnd];
            for (i = 0; i < _n_hydbnd; i++) _hydbnd[i] = tmp[i];
            for (i = _n_hydbnd; i < _alloc_hydbnd; i++) _hydbnd[i] = NULL;
            delete [] tmp;
       }
       _hydbnd[_n_hydbnd] = new _HYDBND;
       strcpy(_hydbnd[_n_hydbnd]->FstAtom, FstAtom);
       strcpy(_hydbnd[_n_hydbnd]->altLoc1, altLoc1);
       _hydbnd[_n_hydbnd]->FstIndex = FstIndex;
       _hydbnd[_n_hydbnd]->FstRes = FstRes;
       strcpy(_hydbnd[_n_hydbnd]->H_Atom, H_Atom);
       strcpy(_hydbnd[_n_hydbnd]->altLocH, altLocH);
       _hydbnd[_n_hydbnd]->H_Index = H_Index;
       _hydbnd[_n_hydbnd]->H_Res = H_Res;
       strcpy(_hydbnd[_n_hydbnd]->SndAtom, SndAtom);
       strcpy(_hydbnd[_n_hydbnd]->altLoc2, altLoc2);
       _hydbnd[_n_hydbnd]->SndIndex = SndIndex;
       _hydbnd[_n_hydbnd]->SndRes = SndRes;
       strcpy(_hydbnd[_n_hydbnd]->SymOP_1, SymOP_1);
       strcpy(_hydbnd[_n_hydbnd]->SymOP_2, SymOP_2);
       _n_hydbnd++;
}

void _FileObj::Insert_a_Link(const char *FstAtom, const char *altLoc1, const int FstIndex,
           const int FstRes, const char *SndAtom, const char *altLoc2, const int SndIndex,
           const int SndRes, const char *SymOP_1, const char *SymOP_2, const char *details)
{
       int i = 0;
       for (i = 0; i < _n_link; i++) {
            if (!strcmp(_link[i]->FstAtom, FstAtom) && !strcmp(_link[i]->altLoc1, altLoc1) &&
                _link[i]->FstIndex == FstIndex && _link[i]->FstRes == FstRes &&
                !strcmp(_link[i]->SndAtom, SndAtom) && !strcmp(_link[i]->altLoc2, altLoc2) &&
                _link[i]->SndIndex == SndIndex && _link[i]->SndRes == SndRes &&
                !strcmp(_link[i]->SymOP_1, SymOP_1) && !strcmp(_link[i]->SymOP_2, SymOP_2) ||
                !strcmp(_link[i]->FstAtom, SndAtom) && !strcmp(_link[i]->altLoc1, altLoc2) &&
                _link[i]->FstIndex == SndIndex && _link[i]->FstRes == SndRes &&
                !strcmp(_link[i]->SndAtom, FstAtom) && !strcmp(_link[i]->altLoc2, altLoc1) &&
                _link[i]->SndIndex == FstIndex && _link[i]->SndRes == FstRes &&
                !strcmp(_link[i]->SymOP_1, SymOP_2) && !strcmp(_link[i]->SymOP_2, SymOP_1))
                 break;
       }
       if (i < _n_link) return;

       if (_n_link >= _alloc_link) {
            _alloc_link += 10;
            _LINK **tmp = _link;
            _link = new _LINK*[_alloc_link];
            for (i = 0; i < _n_link; i++) _link[i] = tmp[i];
            for (i = _n_link; i < _alloc_link; i++) _link[i] = NULL;
            delete [] tmp;
       }
       _link[_n_link] = new _LINK;
       strcpy(_link[_n_link]->FstAtom, FstAtom);
       strcpy(_link[_n_link]->altLoc1, altLoc1);
       _link[_n_link]->FstIndex = FstIndex;
       _link[_n_link]->FstRes = FstRes;
       strcpy(_link[_n_link]->SndAtom, SndAtom);
       strcpy(_link[_n_link]->altLoc2, altLoc2);
       _link[_n_link]->SndIndex = SndIndex;
       _link[_n_link]->SndRes = SndRes;
       strcpy(_link[_n_link]->SymOP_1, SymOP_1);
       strcpy(_link[_n_link]->SymOP_2, SymOP_2);
       int len = strlen(details) + 1;
       _link[_n_link]->details = new char[len];
       strcpy(_link[_n_link]->details, details);
       _n_link++;
}

void _FileObj::Insert_a_Sltbrg(const char *FstAtom, const char *altLoc1, const int FstIndex,
           const int FstRes, const char *SndAtom, const char *altLoc2, const int SndIndex,
           const int SndRes, const char *SymOP_1, const char *SymOP_2)
{
       int i = 0;
       for (i = 0; i < _n_sltbrg; i++) {
            if (!strcmp(_sltbrg[i]->FstAtom, FstAtom) &&
                !strcmp(_sltbrg[i]->altLoc1, altLoc1) &&
                _sltbrg[i]->FstIndex == FstIndex && _sltbrg[i]->FstRes == FstRes &&
                !strcmp(_sltbrg[i]->SndAtom, SndAtom) &&
                !strcmp(_sltbrg[i]->altLoc2, altLoc2) &&
                _sltbrg[i]->SndIndex == SndIndex && _sltbrg[i]->SndRes == SndRes &&
                !strcmp(_sltbrg[i]->SymOP_1, SymOP_1) &&
                !strcmp(_sltbrg[i]->SymOP_2, SymOP_2)) break;
       }
       if (i < _n_sltbrg) return;

       if (_n_sltbrg >= _alloc_sltbrg) {
            _alloc_sltbrg += 5;
            _SLTBRG** tmp = _sltbrg;
            _sltbrg = new _SLTBRG*[_alloc_sltbrg];
            for (i = 0; i < _n_sltbrg; i++) _sltbrg[i] = tmp[i];
            for (i = _n_sltbrg; i < _alloc_sltbrg; i++) _sltbrg[i] = NULL;
            delete [] tmp;
       }
       _sltbrg[_n_sltbrg] = new _SLTBRG;
       strcpy(_sltbrg[_n_sltbrg]->FstAtom, FstAtom);
       strcpy(_sltbrg[_n_sltbrg]->altLoc1, altLoc1);
       _sltbrg[_n_sltbrg]->FstIndex = FstIndex;
       _sltbrg[_n_sltbrg]->FstRes = FstRes;
       strcpy(_sltbrg[_n_sltbrg]->SndAtom, SndAtom);
       strcpy(_sltbrg[_n_sltbrg]->altLoc2, altLoc2);
       _sltbrg[_n_sltbrg]->SndIndex = SndIndex;
       _sltbrg[_n_sltbrg]->SndRes = SndRes;
       strcpy(_sltbrg[_n_sltbrg]->SymOP_1, SymOP_1);
       strcpy(_sltbrg[_n_sltbrg]->SymOP_2, SymOP_2);
       _sltbrg[_n_sltbrg]->details = NULL;
       _n_sltbrg++;
}

void _FileObj::Insert_a_Site(const char *SiteID, const int initIndex, const int initRes)
{
       int i = 0, j = 0;
       for (i = 0; i < _n_site; i++) {
            if (!strcmp(SiteID, _site[i]->SiteID)) break;
       }
       if (i == _n_site) {
            if (_n_site >= _alloc_site) {
                 _alloc_site += 10;
                 _SITE** tmp = _site;
                 _site = new _SITE*[_alloc_site];
                 for (j = 0; j < _n_site; j++) _site[j] = tmp[j];
                 for (j = _n_site; j < _alloc_site; j++) _site[j] = NULL;
                 delete [] tmp;
            }
            _site[_n_site] = new _SITE;
            _n_site++;
            strcpy(_site[i]->SiteID, SiteID);
            _site[i]->numRes = 0;
            _site[i]->alloc_numRes = 5;
            _site[i]->initIndex = new int[5];
            _site[i]->initRes = new int[5];
       }
       for (j = 0; j < _site[i]->numRes; j++) {
            if (_site[i]->initIndex[j] == initIndex &&
                _site[i]->initRes[j] == initRes) break;
       }
       if (j == _site[i]->numRes) {
            if (_site[i]->numRes >= _site[i]->alloc_numRes) {
                 _site[i]->alloc_numRes += 5;
                 int *tmp = _site[i]->initIndex;
                 _site[i]->initIndex = new int[_site[i]->alloc_numRes];
                 for (j = 0; j < _site[i]->numRes; j++)
                      _site[i]->initIndex[j] = tmp[j];
                 delete [] tmp;
                 tmp = _site[i]->initRes;
                 _site[i]->initRes = new int[_site[i]->alloc_numRes];
                 for (j = 0; j < _site[i]->numRes; j++)
                      _site[i]->initRes[j] = tmp[j];
                 delete [] tmp;
            }
            _site[i]->initIndex[_site[i]->numRes] = initIndex;
            _site[i]->initRes[_site[i]->numRes] = initRes;
            _site[i]->numRes++;
       }
}

void _FileObj::Insert_a_Het(const int initIndex, const int initRes, const int numAtoms)
{
       int i = 0;
       for (i = 0; i < _n_het; i++) {
            if (_het[i]->initIndex == initIndex && _het[i]->initRes == initRes) break;
       }
       if (i < _n_het) return;

       if (_n_het >= _alloc_het) {
            _alloc_het += 5;
            _HET** tmp = _het;
            _het = new _HET*[_alloc_het];
            for (i = 0; i < _n_het; i++) _het[i] = tmp[i];
            for (i = _n_het; i < _alloc_het; i++) _het[i] = NULL;
            delete [] tmp;
       }
       _het[_n_het] = new _HET;
       _het[_n_het]->initIndex = initIndex;
       _het[_n_het]->initRes = initRes;
       _het[_n_het]->numAtoms = numAtoms;
       _n_het++;
}

void _FileObj::Insert_a_Modres(const char *PDB_ID, const int FstIndex, const int FstRes,
            const char *Standard_Name, const char *details)
{
       int i = 0;
       for (i = 0; i < _n_modres; i++) {
            if (!strcmp(_modres[i]->PDB_ID, PDB_ID) && _modres[i]->FstIndex == FstIndex &&
                 _modres[i]->FstRes == FstRes) break;
       }
       if (i < _n_modres) {
            if (!strcmp(_modres[i]->Standard_Name, "") && Standard_Name &&
                 strcmp(Standard_Name, ""))
                 strcpy(_modres[i]->Standard_Name, Standard_Name);
            return;
       }

       if (_n_modres >= _alloc_modres) {
            _alloc_modres += 5;
            _MODRES** tmp = _modres;
            _modres = new _MODRES*[_alloc_modres];
            for (i = 0; i < _n_modres; i++) _modres[i] = tmp[i];
            for (i = _n_modres; i < _alloc_modres; i++) _modres[i] = NULL;
            delete [] tmp;
       }
       _modres[_n_modres] = new _MODRES;
       strcpy(_modres[_n_modres]->PDB_ID, PDB_ID);
       _modres[_n_modres]->FstIndex = FstIndex;
       _modres[_n_modres]->FstRes = FstRes;
       if (Standard_Name && strcmp(Standard_Name, ""))
            strcpy(_modres[_n_modres]->Standard_Name, Standard_Name);
       else strcpy(_modres[_n_modres]->Standard_Name, "");
       int len = strlen(details) + 1;
       _modres[_n_modres]->details = new char[len];
       strcpy(_modres[_n_modres]->details, details);
       _n_modres++;
}

void _FileObj::output_structural_features(const int chain_flag)
{
       if (_num_mol == 0 || _molecules == NULL) return;

       _MOLECULE *mol = _molecules[0];
       int serial_helix = 0;
       int serial_turn = 0;
       int i = 0, j = 0, length = 0;
       One_Card_Info_Format *pCard = NULL;
       _HELIX *helix = NULL;
       _TURN *turn = NULL;
       _FIELD *Init = NULL, *End = NULL, *Cur = NULL, *Prev = NULL;
       _CHAIN *chain = mol->GetFirstChain();
       while (chain) {
            for (j = 0; j < chain->n_helix(); j++) {
                 pCard = _CardInfo[HELIX_TOKEN-1].
                         add_a_new_card(NdbToken[HELIX_TOKEN-1], 0);
                 serial_helix++;
                 pCard->AddElementToField(serial_helix, 1);
                 helix = chain->helix(j);
                 pCard->AddElementToField(helix->helixID, 2);
                 pCard->AddElementToField(helix->helixClass, 11);
                 pCard->AddElementToField(helix->comment, 12);
                 length = helix->endRes - helix->initRes + 1;
                 pCard->AddElementToField(length, 13);
                 Init = chain->SeqResIndex(helix->initRes);
                 End  = chain->SeqResIndex(helix->endRes);
                 if (chain_flag == 2) {
                      pCard->AddElementToField(Init->Field[5], 3);
                      pCard->AddElementToField(chain->PUB_ChainID(), 4);
                      pCard->AddElementToField(Init->Field[6], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      pCard->AddElementToField(End->Field[5], 7);
                      pCard->AddElementToField(chain->PUB_ChainID(), 8);
                      pCard->AddElementToField(End->Field[6], 9);
                      pCard->AddElementToField(End->InsCode, 10);
                 } else {
                      pCard->AddElementToField(Init->Field[3], 3);
                      pCard->AddElementToField(Init->Field[4], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      pCard->AddElementToField(End->Field[3], 7);
                      pCard->AddElementToField(End->Field[4], 9);
                      pCard->AddElementToField(End->InsCode, 10);
                      if (chain_flag) {
                           pCard->AddElementToField("", 4);
                           pCard->AddElementToField("", 8);
                      } else {
                           pCard->AddElementToField(chain->PDB_ChainID(), 4);
                           pCard->AddElementToField(chain->PDB_ChainID(), 8);
                      }
                 }
            }

            for (j = 0; j < chain->n_turn(); j++) {
                 pCard = _CardInfo[TURN_TOKEN-1].
                         add_a_new_card(NdbToken[TURN_TOKEN-1], 0);
                 serial_turn++;
                 pCard->AddElementToField(serial_turn, 1);
                 turn = chain->turn(j);
                 pCard->AddElementToField(turn->turnID, 2);
                 pCard->AddElementToField(turn->comment, 11);
                 Init = chain->SeqResIndex(turn->initRes);
                 End  = chain->SeqResIndex(turn->endRes);
                 if (chain_flag == 2) {
                      pCard->AddElementToField(Init->Field[5], 3);
                      pCard->AddElementToField(chain->PUB_ChainID(), 4);
                      pCard->AddElementToField(Init->Field[6], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      pCard->AddElementToField(End->Field[5], 7);
                      pCard->AddElementToField(chain->PUB_ChainID(), 8);
                      pCard->AddElementToField(End->Field[6], 9);
                      pCard->AddElementToField(End->InsCode, 10);
                 } else {
                      pCard->AddElementToField(Init->Field[3], 3);
                      pCard->AddElementToField(Init->Field[4], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      pCard->AddElementToField(End->Field[3], 7);
                      pCard->AddElementToField(End->Field[4], 9);
                      pCard->AddElementToField(End->InsCode, 10);
                      if (chain_flag) {
                           pCard->AddElementToField("", 4);
                           pCard->AddElementToField("", 8);
                      } else {
                           pCard->AddElementToField(chain->PDB_ChainID(), 4);
                           pCard->AddElementToField(chain->PDB_ChainID(), 8);
                      }
                 }
            }
            chain = mol->GetNextChain();
       }

       char sheetID[4];
       strcpy(sheetID, "xxx");
       int serial_sheet = 1;
       _CHAIN *ch1 = NULL, *ch2 = NULL, *ch3 = NULL, *ch4 = NULL;
       for (i = 0; i < _n_sheet; i++) {
            ch1 = mol->GetIndexChain(_sheet[i]->initIndex);
            ch2 = mol->GetIndexChain(_sheet[i]->endIndex);
            ch3 = mol->GetIndexChain(_sheet[i]->curIndex);
            ch4 = mol->GetIndexChain(_sheet[i]->prevIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_sheet[i]->initRes);
            End  = ch2->SeqResIndex(_sheet[i]->endRes);
            if (!Init || !End) continue;

            Cur = NULL; Prev = NULL;
            if (ch3) Cur  = ch3->SeqResIndex(_sheet[i]->curRes);
            if (ch4) Prev = ch4->SeqResIndex(_sheet[i]->prevRes);

            pCard = _CardInfo[SHEET_TOKEN-1].
                    add_a_new_card(NdbToken[SHEET_TOKEN-1], 0);
            if (strcmp(_sheet[i]->sheetID, sheetID)) {
                 strcpy(sheetID, _sheet[i]->sheetID);
                 serial_sheet = 1;
            } else serial_sheet++;
            pCard->AddElementToField(serial_sheet, 1);
            pCard->AddElementToField(_sheet[i]->sheetID, 2);
            pCard->AddElementToField(_sheet[i]->numStrands, 3);
            pCard->AddElementToField(_sheet[i]->sense, 12);
            if (Cur && Prev) {
                 pCard->AddElementToField(_sheet[i]->curAtom, 13);
                 pCard->AddElementToField(_sheet[i]->prevAtom, 18);
            }
            if (chain_flag == 2) {
                 pCard->AddElementToField(Init->Field[5], 4);
                 pCard->AddElementToField(ch1->PUB_ChainID(), 5);
                 pCard->AddElementToField(Init->Field[6], 6);
                 pCard->AddElementToField(Init->InsCode, 7);
                 pCard->AddElementToField(End->Field[5], 8);
                 pCard->AddElementToField(ch2->PUB_ChainID(), 9);
                 pCard->AddElementToField(End->Field[6], 10);
                 pCard->AddElementToField(End->InsCode, 11);
                 if (Cur && Prev) {
                      pCard->AddElementToField(Cur->Field[5], 14);
                      pCard->AddElementToField(ch3->PUB_ChainID(), 15);
                      pCard->AddElementToField(Cur->Field[6], 16);
                      pCard->AddElementToField(Cur->InsCode, 17);
                      pCard->AddElementToField(Prev->Field[5], 19);
                      pCard->AddElementToField(ch4->PUB_ChainID(), 20);
                      pCard->AddElementToField(Prev->Field[6], 21);
                      pCard->AddElementToField(Prev->InsCode, 22);
                 }
            } else {
                 pCard->AddElementToField(Init->Field[3], 4);
                 pCard->AddElementToField(Init->Field[4], 6);
                 pCard->AddElementToField(Init->InsCode, 7);
                 pCard->AddElementToField(End->Field[3], 8);
                 pCard->AddElementToField(End->Field[4], 10);
                 pCard->AddElementToField(End->InsCode, 11);
                 if (chain_flag) {
                      pCard->AddElementToField("", 5);
                      pCard->AddElementToField("", 9);
                 } else {
                      pCard->AddElementToField(ch1->PDB_ChainID(), 5);
                      pCard->AddElementToField(ch2->PDB_ChainID(), 9);
                 }
                 if (Cur && Prev) {
                      pCard->AddElementToField(Cur->Field[3], 14);
                      pCard->AddElementToField(Cur->Field[4], 16);
                      pCard->AddElementToField(Cur->InsCode, 17);
                      pCard->AddElementToField(Prev->Field[3], 19);
                      pCard->AddElementToField(Prev->Field[4], 21);
                      pCard->AddElementToField(Prev->InsCode, 22);
                      if (chain_flag) {
                           pCard->AddElementToField("", 15);
                           pCard->AddElementToField("", 20);
                      } else {
                           pCard->AddElementToField(ch3->PDB_ChainID(), 15);
                           pCard->AddElementToField(ch4->PDB_ChainID(), 20);
                      }
                 }
            }
       }

       int serial_no = 0;
       for (i = 0; i < _n_ssbond; i++) {
            ch1 = mol->GetIndexChain(_ssbond[i]->FstIndex);
            ch2 = mol->GetIndexChain(_ssbond[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_ssbond[i]->FstRes);
            End  = ch2->SeqResIndex(_ssbond[i]->SndRes);
            if (Init && End) {
                 pCard = _CardInfo[SSBOND_TOKEN-1].
                         add_a_new_card(NdbToken[SSBOND_TOKEN-1], 0);
                 serial_no++;
                 pCard->AddElementToField(serial_no, 1);
                 pCard->AddElementToField(_ssbond[i]->SymOP_1, 10);
                 pCard->AddElementToField(_ssbond[i]->SymOP_2, 11);
                 if (chain_flag == 2) {
                      pCard->AddElementToField(Init->Field[5], 2);
                      pCard->AddElementToField(ch1->PUB_ChainID(), 3);
                      pCard->AddElementToField(Init->Field[6], 4);
                      pCard->AddElementToField(Init->InsCode, 5);
                      pCard->AddElementToField(End->Field[5], 6);
                      pCard->AddElementToField(ch2->PUB_ChainID(), 7);
                      pCard->AddElementToField(End->Field[6], 8);
                      pCard->AddElementToField(End->InsCode, 9);
                 } else {
                      pCard->AddElementToField(Init->Field[3], 2);
                      pCard->AddElementToField(Init->Field[4], 4);
                      pCard->AddElementToField(Init->InsCode, 5);
                      pCard->AddElementToField(End->Field[3], 6);
                      pCard->AddElementToField(End->Field[4], 8);
                      pCard->AddElementToField(End->InsCode, 9);
                      if (chain_flag) {
                           pCard->AddElementToField("", 3);
                           pCard->AddElementToField("", 7);
                      } else {
                           pCard->AddElementToField(ch1->PDB_ChainID(), 3);
                           pCard->AddElementToField(ch2->PDB_ChainID(), 7);
                      }
                 }
            }
       }

       for (i = 0; i < _n_link; i++) {
            ch1 = mol->GetIndexChain(_link[i]->FstIndex);
            ch2 = mol->GetIndexChain(_link[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_link[i]->FstRes);
            End  = ch2->SeqResIndex(_link[i]->SndRes);
            if (Init && End) {
                 pCard = _CardInfo[LINK_TOKEN-1].
                         add_a_new_card(NdbToken[LINK_TOKEN-1], 0);
                 pCard->AddElementToField(_link[i]->FstAtom, 1);
                 pCard->AddElementToField(_link[i]->altLoc1, 2);
                 pCard->AddElementToField(Init->InsCode, 6);
                 pCard->AddElementToField(_link[i]->SndAtom, 7);
                 pCard->AddElementToField(_link[i]->altLoc2, 8);
                 pCard->AddElementToField(End->InsCode, 12);
                 pCard->AddElementToField(_link[i]->SymOP_1, 13);
                 pCard->AddElementToField(_link[i]->SymOP_2, 14);
                 if (chain_flag == 2) {
                      pCard->AddElementToField(Init->Field[5], 3);
                      pCard->AddElementToField(ch1->PUB_ChainID(), 4);
                      pCard->AddElementToField(Init->Field[6], 5);
                      pCard->AddElementToField(End->Field[5], 9);
                      pCard->AddElementToField(ch2->PUB_ChainID(), 10);
                      pCard->AddElementToField(End->Field[6], 11);
                 } else {
                      pCard->AddElementToField(Init->Field[3], 3);
                      pCard->AddElementToField(Init->Field[4], 5);
                      pCard->AddElementToField(End->Field[3], 9);
                      pCard->AddElementToField(End->Field[4], 11);
                      if (chain_flag) {
                           pCard->AddElementToField("", 4);
                           pCard->AddElementToField("", 10);
                      } else {
                           pCard->AddElementToField(ch1->PDB_ChainID(), 4);
                           pCard->AddElementToField(ch2->PDB_ChainID(), 10);
                      }
                 }
            }
       }

       for (i = 0; i < _n_hydbnd; i++) {
            ch1 = mol->GetIndexChain(_hydbnd[i]->FstIndex);
            ch2 = mol->GetIndexChain(_hydbnd[i]->SndIndex);
            ch3 = mol->GetIndexChain(_hydbnd[i]->H_Index);
            if (!ch1 || ch2) continue;
            Init = ch1->SeqResIndex(_hydbnd[i]->FstRes);
            End  = ch2->SeqResIndex(_hydbnd[i]->SndRes);
            Cur  = NULL;
            if (ch3) Cur = ch3->SeqResIndex(_hydbnd[i]->H_Res);
            if (Init && End) {
                 pCard = _CardInfo[HYDBND_TOKEN-1].
                         add_a_new_card(NdbToken[HYDBND_TOKEN-1], 0);
                 pCard->AddElementToField(_hydbnd[i]->FstAtom, 1);
                 pCard->AddElementToField(_hydbnd[i]->altLoc1, 2);
                 pCard->AddElementToField(_hydbnd[i]->H_Atom, 7);
                 pCard->AddElementToField(_hydbnd[i]->altLocH, 8);
                 pCard->AddElementToField(_hydbnd[i]->SndAtom, 12);
                 pCard->AddElementToField(_hydbnd[i]->altLoc2, 13);
                 pCard->AddElementToField(_hydbnd[i]->SymOP_1, 18);
                 pCard->AddElementToField(_hydbnd[i]->SymOP_2, 19);
                 if (chain_flag == 2) {
                      pCard->AddElementToField(Init->Field[5], 3);
                      pCard->AddElementToField(ch1->PUB_ChainID(), 4);
                      pCard->AddElementToField(Init->Field[6], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      if (Cur) {
                           pCard->AddElementToField(ch3->PUB_ChainID(), 10);
                           pCard->AddElementToField(Cur->Field[6], 11);
                           pCard->AddElementToField(Cur->InsCode, 12);
                      }
                      pCard->AddElementToField(End->Field[5], 14);
                      pCard->AddElementToField(ch2->PUB_ChainID(), 15);
                      pCard->AddElementToField(End->Field[6], 16);
                      pCard->AddElementToField(End->InsCode, 17);
                 } else {
                      pCard->AddElementToField(Init->Field[3], 3);
                      pCard->AddElementToField(Init->Field[4], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      if (Cur) {
                           if (chain_flag)
                                pCard->AddElementToField(ch3->PDB_ChainID(), 10);
                           else pCard->AddElementToField("", 10);
                           pCard->AddElementToField(Cur->Field[4], 11);
                           pCard->AddElementToField(Cur->InsCode, 12);
                      }
                      pCard->AddElementToField(End->Field[3], 14);
                      pCard->AddElementToField(End->Field[4], 16);
                      pCard->AddElementToField(End->InsCode, 17);
                      if (chain_flag) {
                           pCard->AddElementToField("", 4);
                           pCard->AddElementToField("", 15);
                      } else {
                           pCard->AddElementToField(ch1->PDB_ChainID(), 4);
                           pCard->AddElementToField(ch2->PDB_ChainID(), 15);
                      }
                 }
            }
       }

       for (i = 0; i < _n_sltbrg; i++) {
            ch1 = mol->GetIndexChain(_sltbrg[i]->FstIndex);
            ch2 = mol->GetIndexChain(_sltbrg[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_sltbrg[i]->FstRes);
            End  = ch2->SeqResIndex(_sltbrg[i]->SndRes);
            if (Init && End) {
                 pCard = _CardInfo[SLTBRG_TOKEN-1].
                         add_a_new_card(NdbToken[SLTBRG_TOKEN-1], 0);
                 pCard->AddElementToField(_sltbrg[i]->FstAtom, 1);
                 pCard->AddElementToField(_sltbrg[i]->altLoc1, 2);
                 pCard->AddElementToField(_sltbrg[i]->SndAtom, 7);
                 pCard->AddElementToField(_sltbrg[i]->altLoc2, 8);
                 pCard->AddElementToField(_sltbrg[i]->SymOP_1, 13);
                 pCard->AddElementToField(_sltbrg[i]->SymOP_2, 14);
                 if (chain_flag == 2) {
                      pCard->AddElementToField(Init->Field[5], 3);
                      pCard->AddElementToField(ch1->PUB_ChainID(), 4);
                      pCard->AddElementToField(Init->Field[6], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      pCard->AddElementToField(End->Field[5], 9);
                      pCard->AddElementToField(ch2->PUB_ChainID(), 10);
                      pCard->AddElementToField(End->Field[6], 11);
                      pCard->AddElementToField(End->InsCode, 12);
                 } else {
                      pCard->AddElementToField(Init->Field[3], 3);
                      pCard->AddElementToField(Init->Field[4], 5);
                      pCard->AddElementToField(Init->InsCode, 6);
                      pCard->AddElementToField(End->Field[3], 9);
                      pCard->AddElementToField(End->Field[4], 11);
                      pCard->AddElementToField(End->InsCode, 12);
                      if (chain_flag) {
                           pCard->AddElementToField("", 4);
                           pCard->AddElementToField("", 10);
                      } else {
                           pCard->AddElementToField(ch1->PDB_ChainID(), 4);
                           pCard->AddElementToField(ch2->PDB_ChainID(), 10);
                      }
                 }
            }
       }

       serial_no = 0;
       char SiteID[10];
       strcpy(SiteID, "XXXXX");
       int siteId = 0;
       for (i = 0; i < _n_site; i++) {
            siteId = 0;
            pCard = NULL;
            for (j = 0; j < _site[i]->numRes; j++) {
                 ch1 = mol->GetIndexChain(_site[i]->initIndex[j]);
                 if (!ch1) continue;
                 Init = ch1->SeqResIndex(_site[i]->initRes[j]);
                 if (!Init) continue;

                 if (strcmp(SiteID, _site[i]->SiteID)) {
                      serial_no = 0;
                      strcpy(SiteID, _site[i]->SiteID);
                 }

                 if (siteId == 0) {
                      pCard = _CardInfo[SITE_TOKEN-1].
                              add_a_new_card(NdbToken[SITE_TOKEN-1], 0);
                      serial_no++;
                      pCard->AddElementToField(serial_no, 1);
                      pCard->AddElementToField(_site[i]->SiteID, 2);
                      pCard->AddElementToField(_site[i]->numRes, 3);
                 }
                 if (chain_flag == 2) {
                      pCard->AddElementToField(Init->Field[5], siteId * 4 + 4);
                      pCard->AddElementToField(ch1->PUB_ChainID(), siteId * 4 + 5);
                      pCard->AddElementToField(Init->Field[6], siteId * 4 + 6);
                      pCard->AddElementToField(Init->InsCode, siteId * 4 + 7);
                 } else {
                      pCard->AddElementToField(Init->Field[3], siteId * 4 + 4);
                      if (chain_flag)
                           pCard->AddElementToField("", siteId * 4 + 5);
                      else pCard->AddElementToField(ch1->PDB_ChainID(), siteId * 4 + 5);
                      pCard->AddElementToField(Init->Field[4], siteId * 4 + 6);
                      pCard->AddElementToField(Init->InsCode, siteId * 4 + 7);
                 }
                 siteId++;
                 if (siteId == 4) siteId = 0;
            }
       }
 
       serial_no = 0;
       for (i = 0; i < _n_modres; i++) {
            ch1 = mol->GetIndexChain(_modres[i]->FstIndex);
            if (!ch1) continue;
            Init = ch1->SeqResIndex(_modres[i]->FstRes);
            if (!Init) continue;

            pCard = _CardInfo[MODRES_TOKEN-1].
                    add_a_new_card(NdbToken[MODRES_TOKEN-1], 0);
            pCard->AddElementToField(_modres[i]->PDB_ID, 1);
            pCard->AddElementToField(_modres[i]->Standard_Name, 6);
            pCard->AddElementToField(_modres[i]->details, 7);
            serial_no++;
            pCard->AddElementToField(serial_no, 8);
            if (chain_flag == 2) {
                 pCard->AddElementToField(Init->Field[5], 2);
                 pCard->AddElementToField(ch1->PUB_ChainID(), 3);
                 pCard->AddElementToField(Init->Field[6], 4);
                 pCard->AddElementToField(Init->InsCode, 5);
            } else {
                 pCard->AddElementToField(Init->Field[3], 2);
                 if (chain_flag)
                      pCard->AddElementToField("", 3);
                 else pCard->AddElementToField(ch1->PDB_ChainID(), 3);
                 pCard->AddElementToField(Init->Field[4], 4);
                 pCard->AddElementToField(Init->InsCode, 5);
            }
       }

       for (i = 0; i < _n_het; i++) {
            ch1 = mol->GetIndexChain(_het[i]->initIndex);
            if (!ch1) continue;
            Init = ch1->SeqResIndex(_het[i]->initRes);
            if (!Init) continue;

            pCard = _CardInfo[HET_TOKEN-1].
                    add_a_new_card(NdbToken[HET_TOKEN-1], 0);
            if (chain_flag == 2) {
                 pCard->AddElementToField(Init->Field[5], 1);
                 pCard->AddElementToField(ch1->PUB_ChainID(), 2);
                 pCard->AddElementToField(Init->Field[6], 3);
                 pCard->AddElementToField(Init->InsCode, 4);
            } else {
                 pCard->AddElementToField(Init->Field[3], 1);
                 if (chain_flag)
                      pCard->AddElementToField("", 2);
                 else pCard->AddElementToField(ch1->PDB_ChainID(), 2);
                 pCard->AddElementToField(Init->Field[4], 3);
                 pCard->AddElementToField(Init->InsCode, 4);
            }
            pCard->AddElementToField(_het[i]->numAtoms, 5);
       }
}

